/**
 * \file
 * \brief SCOP IPv6 module (Core Component)
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <err.h>
#include <errno.h>

#include <fcntl.h>

#include <signal.h>

#include "scop.h"
#include "tools.h"
#include "ipv6.h"

void sigintHandler(int bla)
{
  scopTransportQuit();
  putchar('\n');
  exit(42);
}

void scopRun()
{
  int result;
  struct sigaction sa;

#ifdef TRACE
  DEBUG_indent += 2;
  for(DEBUG_i=0; DEBUG_i<(DEBUG_indent-2); DEBUG_i+=2, printf("| "));
  printf("|-TRACE: %s()\n",__FUNCTION__);
#endif

  sa.sa_handler = sigintHandler;
  sa.sa_flags = SA_RESTART;
  sigprocmask(SIG_SETMASK, &sa.sa_mask,NULL);

  sigaction(SIGINT, &sa, NULL);

  /* initialize transport module */
  scopTransportInit();

  while(1)
    {
      /* read and decode a message */
      result = scopDecode();
      if(result != 0 && global_errmsg != NULL)
	{
	  scopResponse(result, global_errmsg);
	}
    }

  return;
}


int scopIfaceInit(char *tag)
{
  char *iface_name, *iface_mac, *iface_mtu, *iface_ipv6, *route, *dest, *next;
  int i, result;

#ifdef TRACE
  DEBUG_indent += 2;
  for(DEBUG_i=0; DEBUG_i<(DEBUG_indent-2); DEBUG_i+=2, printf("| "));
  printf("|-TRACE: %s()\n",__FUNCTION__);
#endif

  /* get interface name */
  toolGetTag(tag, "name", 1, &iface_name);
  if(iface_name == NULL)
    {
      /* invalid iface config section */
      free(tag);
      warnx("%s(): %s", __FUNCTION__, IPV6_CONFIG_INVALIDIFACE_ERRMSG);
      global_errmsg = IPV6_CONFIG_INVALIDIFACE_ERRMSG;
      return IPV6_CONFIG_INVALIDIFACE_ERRCODE;
    }

  /* shut down iface */
  result = ipv6IfaceStop(iface_name);
  if(result != 0)
    {
      warnx("%s(): %s", __FUNCTION__, IPV6_IFACESTOP_ERRMSG);
      global_errmsg = IPV6_IFACESTOP_ERRMSG;
      return result;
    }

  /* set MAC address */
  toolGetTag(tag, "mac_addr", 1, &iface_mac);
  result = ipv6IfaceSetMac(iface_name, iface_mac);
  if(result != 0)
    {
      warnx("%s(): %s", __FUNCTION__, IPV6_IFACEMAC_ERRMSG);
      free(iface_name);
      free(iface_mac);
      global_errmsg = IPV6_IFACEMAC_ERRMSG;
      return result;
    }

  free(iface_mac);

  /* set MTU */
  toolGetTag(tag, "mtu", 1, &iface_mtu);
  result = ipv6IfaceSetMtu(iface_name, iface_mtu);
  if(result != 0)
    {
      warnx("%s(): %s", __FUNCTION__, IPV6_IFACEMTU_ERRMSG);
      free(iface_name);
      free(iface_mtu);
      global_errmsg = IPV6_IFACEMTU_ERRMSG;
      return result;
    }

  free(iface_mtu);

  i = 0;
  do
    {
      i++;
      
      /* set IPv6 address */
      toolGetTag(tag, "ipv6_addr", i, &iface_ipv6);
      if(iface_ipv6 != NULL)
	{
	  result = ipv6IfaceSetAddr(iface_name, iface_ipv6);
	  if(result != 0)
	    {
	      warnx("%s(): %s", __FUNCTION__, IPV6_IFACEADDR_ERRMSG);
	      free(iface_name);
	      free(iface_ipv6);
	      global_errmsg = IPV6_IFACEADDR_ERRMSG;
	      return result;
	    }

	  free(iface_ipv6);
	}
    }
  while(iface_ipv6 != NULL);

  /* activate iface */
  result = ipv6IfaceStart(iface_name);
  if(result != 0)
    {
      warnx("%s(): %s", __FUNCTION__, IPV6_IFACESTART_ERRMSG);
      free(iface_name);
      global_errmsg = IPV6_IFACESTART_ERRMSG;
      return result;
    }

  i = 0;
  do
    {
      i++;
      
      /* set route */
      toolGetTag(tag, "route", i, &route);
      if(route != NULL)
	{
	  toolGetTag(route, "destination", 1, &dest);
	  toolGetTag(route, "nexthop", 1, &next);
	  if( (dest != NULL) && (next != NULL) )
	    {
	      result = ipv6IfaceSetRoute(dest, next);
	      if(result != 0)
		{
		  warnx("%s(): %s", __FUNCTION__, IPV6_IFACEADDR_ERRMSG);
		  free(iface_name);
		  free(route);
		  free(dest);
		  free(next);
		  global_errmsg = IPV6_IFACEROUTE_ERRMSG;
		  return result;
		}
	    }
	  free(route);
	  free(dest);
	  free(next);
	}
    }
  while(route != NULL);


  free(iface_name);
  return 0;
}

int scopDisableIfaces()
{
  int fd, result, i;
  char *buffer, *tag;

#ifdef TRACE
  DEBUG_indent += 2;
  for(DEBUG_i=0; DEBUG_i<(DEBUG_indent-2); DEBUG_i+=2, printf("| "));
  printf("|-TRACE: %s()\n",__FUNCTION__);
#endif
  
  /* open general config file */
  fd = open(SCOP_CONFIG_FILE , O_RDONLY);
  if(fd == -1)
    {
      warnx("%s(): %s", __FUNCTION__, strerror(errno));
      global_errmsg = strerror(errno);
      return SCOP_EXTERNAL_ERRCODE;
    }

  /* load config file in memory */
  result = toolLoadFile(fd, &buffer);
  if(result != 0)
    {
      warnx("%s(): %s", __FUNCTION__, SCOP_CONFIG_LOADFILE_ERRMSG);
      global_errmsg = SCOP_CONFIG_LOADFILE_ERRMSG;
      return result;
    }	      

  close(fd);

  /* get each interface */
  i = 0;
  do
    {
      i++;
      
      /* get interface name */
      toolGetTag(buffer, "iface", i, &tag);
      if(tag != NULL)
	{
	  /* disable interface */
	  result = ipv6IfaceStop(tag);
	  free(tag);
	  if(result != 0)
	    {
	      free(buffer);
	      warnx("%s(): %s", __FUNCTION__, strerror(errno));
	      global_errmsg = IPV6_IFACESTOP_ERRMSG;
	      return result;
	    }	      
	}
    }
  while(tag != NULL);

  free(buffer);
  
  return 0;
}


int scopInitRequest(char * tc_id)
{
  int fd, result;
  char *filename, *buffer, *tag;

#ifdef TRACE
  DEBUG_indent += 2;
  for(DEBUG_i=0; DEBUG_i<(DEBUG_indent-2); DEBUG_i+=2, printf("| "));
  printf("|-TRACE: %s()\n",__FUNCTION__);
#endif
  
  /* compute config file name for current test-case*/
  /* 7 = '\0' + '/' + ".conf" */
  filename = (char *)malloc(strlen(SCOP_CONFIG_DIR)+strlen(tc_id)+7);
  if(filename==NULL)
    {
      warnx("%s(): %s", __FUNCTION__, strerror(errno));
      scopResponse(SCOP_EXTERNAL_ERRCODE, strerror(errno));
      global_errmsg = NULL;	    
      return SCOP_EXTERNAL_ERRCODE;
    }

  sprintf(filename, "%s/%s.conf", SCOP_CONFIG_DIR, tc_id);
 
  /* open config file */
  fd = open( filename , O_RDONLY);
  if(fd == -1)
    {
      warnx("%s(): %s", __FUNCTION__, strerror(errno));
      free(filename);
      scopResponse(SCOP_EXTERNAL_ERRCODE, strerror(errno));
      global_errmsg = NULL;	    
      return SCOP_EXTERNAL_ERRCODE;
    }

  free(filename);

  /* load config file into memory */
  toolLoadFile(fd, &buffer);
  if(result != 0)
    {
      warnx("%s(): %s", __FUNCTION__, SCOP_CONFIG_LOADFILE_ERRMSG);
      global_errmsg = SCOP_CONFIG_LOADFILE_ERRMSG;
      return result;
    }	      

  close(fd);

  /* disable all interfaces */
  result = scopDisableIfaces();
  if(result != 0)
    {
      free(buffer);
      scopResponse(result, global_errmsg);
      global_errmsg = NULL;
      return result;
    }

  /* flush routes */
  result = ipv6FlushRoutes();
  if(result != 0)
    {
      free(buffer);
      scopResponse(result, global_errmsg);
      global_errmsg = NULL;
      return result;
    }	      

  /* check 'router' tag */ 
  toolGetTag(buffer, "router", 1, &tag);
  if(tag != NULL)
    {
      /* router configuration */
      int i = 0;

      /* activate routing */
      result = ipv6NodeEnableRouter();
      if(result != 0)
	{
	  warnx("%s(): %s", __FUNCTION__, IPV6_ROUTERENABLE_ERRMSG);
	  free(tag);
	  free(buffer);
	  scopResponse(result, IPV6_ROUTERENABLE_ERRMSG);
	  global_errmsg = NULL;
	  return result;
	}

      free(buffer);
      buffer = tag;
      do
	{
	  i++;
	  
	  /* get interface */
	  toolGetTag(buffer, "iface", i, &tag);
	  if(tag != NULL)
	    {
	      /* initialize interface */
	      result = scopIfaceInit(tag);
	      free(tag);
	      if(result != 0)
		{
		  free(buffer);
		  scopResponse(result, global_errmsg);
		  global_errmsg = NULL;
		  return result;
		}	      
	    }
	}
      while(tag != NULL);
    }
  else
    {
      /* check 'host' tag */
      toolGetTag(buffer, "host", 1, &tag);
      if(tag != NULL)
	{	  
	  /* host configuration */
	  int i = 0;
	  
	  /* disable routing */
	  result = ipv6NodeDisableRouter();
	  if(result != 0)
	    {
	      warnx("%s(): %s", __FUNCTION__, IPV6_ROUTERDISABLE_ERRMSG);
	      free(tag);
	      free(buffer);
	      scopResponse(result, IPV6_ROUTERDISABLE_ERRMSG);
	      global_errmsg = NULL;
	      return result;
	    }

	  free(buffer);
	  buffer = tag;
	  do
	    {
	      i++;
	      
	      /* get interface */
	      toolGetTag(buffer, "iface", i, &tag);
	      if(tag != NULL)
		{
		  /* initialize interface */
		  result = scopIfaceInit(tag);
		  free(tag);
		  if(result != 0)
		    {
		      free(buffer);
		      scopResponse(result, global_errmsg);
		      global_errmsg = NULL;
		      return result;
		    }	      
		}
	    }
	  while(tag != NULL);
	}
      else
	{
	  /* check 'node' tag */
	  toolGetTag(buffer, "node", 1, &tag);
	  if(tag != NULL)
	    {	  
	      /* node configuration */
	      int i = 0;

	      free(buffer);
	      buffer = tag;
	      do
		{
		  i++;
	      
		  /* get interface */
		  toolGetTag(buffer, "iface", i, &tag);
		  if(tag != NULL)
		    {
		      /* initialize interface */
		      result = scopIfaceInit(tag);
		      free(tag);
		      if(result != 0)
			{
			  free(buffer);
			  scopResponse(result, global_errmsg);
			  global_errmsg = NULL;
			  return result;
			}	      
		    }
		}
	      while(tag != NULL);
	    }
	  else
	    {
	      /* Ooops : invalid node type */
	      free(buffer);
	      scopResponse(IPV6_CONFIG_INVALIDNODE_ERRCODE, IPV6_CONFIG_INVALIDNODE_ERRMSG);
	      global_errmsg = NULL;	    
	      return IPV6_CONFIG_INVALIDNODE_ERRCODE;
	    }
	}
    }

  free(buffer);
  scopResponse(0, SCOP_OK_ERRMSG);
  return 0;
}

int scopResetRequest()
{

#ifdef TRACE
  DEBUG_indent += 2;
  for(DEBUG_i=0; DEBUG_i<(DEBUG_indent-2); DEBUG_i+=2, printf("| "));
  printf("|-TRACE: %s()\n",__FUNCTION__);
#endif

  /* TODO */

  return 0;
}

int scopDataRequest(char * cmd, unsigned char *data)
{
  int result;
  char *packet;

#ifdef TRACE
  DEBUG_indent += 2;
  for(DEBUG_i=0; DEBUG_i<(DEBUG_indent-2); DEBUG_i+=2, printf("| "));
  printf("|-TRACE: %s()\n",__FUNCTION__);
#endif

  if(strncmp(cmd,SCOP_SEND_VIA_IUT,strlen(SCOP_SEND_VIA_IUT)) == 0)
    {
      result = ipv6Decode(data);

      scopDataResponse(result, NULL);
      global_errmsg = NULL;	    
      return result;
    }
  else
    {
      if(strncmp(cmd,SCOP_GET_VIA_IUT,strlen(SCOP_GET_VIA_IUT)) == 0)
	{
	  /* TODO */
	  result = ipv6Get(&packet);
	  
	  scopDataResponse(result, packet);
	  global_errmsg = NULL;	    
	  return result;
	}
      else
	{
	  /* error (unknown command) */
	  warnx("%s(): %s", __FUNCTION__, SCOP_BADCMD_ERRMSG);
	  scopDataResponse(SCOP_BADCMD_ERRCODE, NULL);
	  global_errmsg = NULL;	    
	  return SCOP_BADCMD_ERRCODE;
	}
    }

  return 0;
}



