/**
 * \file
 * \brief ipv6 module (IUT Interface)
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>

#include <err.h>
#include <errno.h>

#include <fcntl.h>

#include <sys/types.h>
#include <sys/socket.h>

#include <ifaddrs.h>
#include <netdb.h>

#include "ipv6.h"
#include "tools.h"


int ipv6_socket=0;
void DEBUGprintpacket(struct ip6_hdr header, unsigned char *payload, char *tag);

/* dirty */
#include <pcap.h>
pcap_t *sniffer;

void dirty_snif(char * iface, struct in6_addr *src)
{
  char errmsg[PCAP_ERRBUF_SIZE];
  struct pcap_pkthdr header;
  const u_char *packet;
  struct bpf_program filter;
  char filter_string[50] = "src host 0000:0000:0000:0000:0000:0000:0000:0000";
  int i;
  struct ip6_hdr pnt_header;

  pid_t pid;

  /* create a new processus */
  pid = fork();
  if(pid == -1)
    {
      warnx("%s(): %s", __FUNCTION__, strerror(errno));
    }
  if(pid == 0)
    {
      /* new processus */ 

      /* open sniffing device */
      sniffer = pcap_open_live(iface, 100, 1, 2000, errmsg);

      /* set filter, to get only outgoing packets */
      for(i=0; i<16; i+=2)
	sprintf( &(filter_string[strlen("src host ")+(5*i/2)]), "%02X%02X:",
		 ((unsigned char *)src)[i], ((unsigned char *)src)[i+1]);
      filter_string[48]='\0';

      pcap_compile(sniffer, &filter, filter_string, 0, 0);
      pcap_setfilter(sniffer, &filter);

      /* sniff a packet */
/*       pcap_loop(sniffer, 1, dirty_packet, NULL); */
      packet = pcap_next(sniffer, &header);

      memcpy(&pnt_header, (unsigned char*)&(packet[14]), sizeof(pnt_header));

#ifdef DEBUG
      DEBUGprintpacket(pnt_header, (unsigned char*)&(packet[14+sizeof(pnt_header)]), "SENT");
#endif /* DEBUG */

      exit(0);
    }

  sleep(1); 
}


#ifdef __linux__

/* These four functions are missing in Linux IPv6 API           */
/* USAGI IPv6 snap kit usagi-linux26-s20050425.tar.bz2, ping6.c */

/* this is not clean... but I cannot include linux/ipv6.h */
struct ipv6_rt_hdr {
  unsigned char   nexthdr;
  unsigned char   hdrlen;
  unsigned char   type;
  unsigned char   segments_left;
};

/* this is not clean... but I cannot include linux/ipv6.h */
struct rt0_hdr {
  struct ipv6_rt_hdr      rt_hdr;
  unsigned short int      bitmap;         /* strict/loose bit map */
  struct in6_addr         addr[1];
};

size_t inet6_rthdr_space(int type, int segments)
{
  if (type != 0 || segments > 24)
    return 0;

  return (sizeof(struct cmsghdr) + sizeof(struct rt0_hdr) +
	  segments * sizeof(struct in6_addr));
}

extern struct cmsghdr * inet6_rthdr_init(void *bp, int type)
{
  struct cmsghdr *cmsg;

  if (type)
    return NULL;

  memset(bp, 0, sizeof(struct cmsghdr) + sizeof(struct rt0_hdr));
  cmsg = (struct cmsghdr *) bp;

  cmsg->cmsg_len = sizeof(struct cmsghdr) + sizeof(struct rt0_hdr);
  cmsg->cmsg_level = SOL_IPV6;
  cmsg->cmsg_type = IPV6_RTHDR;

  return cmsg;
}

int inet6_rthdr_add(struct cmsghdr *cmsg, const struct in6_addr *addr, unsigned int flags)
{
  struct rt0_hdr *hdr;

  hdr = (struct rt0_hdr *) CMSG_DATA(cmsg);

  cmsg->cmsg_len += sizeof(struct in6_addr);
  hdr->rt_hdr.hdrlen += sizeof(struct in6_addr) / 8;

  memcpy(&(hdr->addr[hdr->rt_hdr.segments_left++]), addr,
	 sizeof(struct in6_addr));

  return 0;
}

int inet6_rthdr_lasthop(struct cmsghdr *cmsg, unsigned int flags)
{
  return 0;
}

#endif /* __linux__ */



int ipv6SendPing(struct ip6_hdr *header, int nb_hop, struct in6_addr *hops)
{
  int i, j, error;
  char **args;
  struct ifaddrs  *ifa, *ifap;
  struct in6_addr *src_addr;

#ifdef TRACE
  DEBUG_indent += 2;
  for(DEBUG_i=0; DEBUG_i<(DEBUG_indent-2); DEBUG_i+=2, printf("| "));
  printf("|-TRACE: %s()\n",__FUNCTION__);
#endif

  /* 9 = 'ping6' + '-c' + '1' + '-h' + <hoplimit> + '-I' + <iface> + <dst_addr> + 'NULL' */
  args = (char **)malloc((nb_hop+9)*sizeof(char *));
  if(args == NULL)
    {
      warnx("%s(): %s", __FUNCTION__, strerror(errno));
      return IPV6_EXTERNAL_ERRCODE;
    }
  
  args[0] = IPV6_PING6_CMD;

  /* set hop limit value */
  args[1] = IPV6_PING6_HLIM_OPT;
  args[2] = (char *)malloc(4); 
  if(args[2] == NULL)
    {
      warnx("%s(): %s", __FUNCTION__, strerror(errno));
      free(args);
      return IPV6_EXTERNAL_ERRCODE;
    }
  snprintf(args[2], 4, "%3d", (int)header->ip6_hlim);

  /* send only one "echo request" */
  args[3] = "-c";
  args[4] = "1";

  /* get configured IPv6 addresses */
  error = getifaddrs(&ifap);
  if(error != 0)
    {
      warnx("%s() %d: %s", __FUNCTION__, __LINE__, strerror(errno));
      return IPV6_EXTERNAL_ERRCODE;
    }
  
  /* find the configured IPv6 address which match src address */
  for(ifa=ifap; ifa; ifa=ifa->ifa_next)
    {
      src_addr = &((struct sockaddr_in6 *)ifa->ifa_addr)->sin6_addr;

      /* FreeBSD link-local addresses are recorded in a strange way */
      if((memcmp(src_addr, &(header->ip6_src), 2) == 0)
	 && (memcmp(((unsigned char *)src_addr) + 4, ((unsigned char *)&(header->ip6_src)) +4,
		    sizeof(struct in6_addr) - 4) == 0))
	break;
    }
  
  if(ifa == NULL)
    {
      warnx("%s(): %s", __FUNCTION__, IPV6_SEND_INVALIDSRC_ERRMSG); 
      freeifaddrs(ifap);
      free(args[2]);
      free(args);
      return IPV6_SEND_INVALIDSRC_ERRCODE;
    }
  
  /* 45 = strlen(ascii_ipv6_addr) + strlen(iface_name) + 1 (%) */
  args[7] = (char *)malloc(45);
  if(args[7] == NULL)
  {
      warnx("%s(): %s", __FUNCTION__, strerror(errno));
      free(args[2]);
      free(args);
      freeifaddrs(ifap);
      return IPV6_EXTERNAL_ERRCODE;
    }

  /* add hops */
  for(j=0; j<14; j+=2)
    sprintf(&(args[7][(5*j)/2]), "%02X%02X:", header->ip6_dst.s6_addr[j], header->ip6_dst.s6_addr[j+1]);
  sprintf(&(args[7][(5*j)/2]), "%02X%02X", header->ip6_dst.s6_addr[j], header->ip6_dst.s6_addr[j+1]);

#ifdef __FreeBSD__

  /* add '%<iface>' to dst if dst is a link-local address */
  if(strncmp(args[7], "FE80", 4) == 0)
    {
      sprintf(&(args[7][39]), "%%%s", ifa->ifa_name);
    }

#endif /* __Freebsd__ */

  if(nb_hop == 0)
    {
      /* set -I <iface> option */
      args[5] = "-I";
      args[6] = ifa->ifa_name;
    }
  else
    {
      args[5] = "-c";
      args[6] = "1";
    }

  for(i=0; i<nb_hop; i++)
    {
      args[8+i] = (char *)malloc(41);
      if(args[8+i] == NULL)
	{
	  warnx("%s(): %s", __FUNCTION__, strerror(errno));
	  free(args[2]);
	  free(args[7]);
	  freeifaddrs(ifap);
	  for(i--;i>=0;i--)
	    free(args[8+i]);
	  return IPV6_EXTERNAL_ERRCODE;
	}
      
      for(j=0; j<14; j+=2)
	sprintf(&(args[8+i][(5*j)/2]), "%02X%02X:", hops[i].s6_addr[j], hops[i].s6_addr[j+1]);
      sprintf(&(args[8+i][(5*j)/2]), "%02X%02X", hops[i].s6_addr[j], hops[i].s6_addr[j+1]);
    }
  
  args[nb_hop+8] = NULL;

  /* execute command */
  error = toolRunProgram(args);

  /* free memory */
  for(i=0; i<nb_hop; i++)
    free(args[8+i]);
  free(args[7]);
  free(args[2]);
  free(args);
  freeifaddrs(ifap);

  return error;
}


int ipv6IfaceStop(char *iface)
{
  int error;
  char *args[3]={IPV6_IFDOWN_CMD, NULL, NULL};

#ifdef TRACE
  DEBUG_indent += 2;
  for(DEBUG_i=0; DEBUG_i<(DEBUG_indent-2); DEBUG_i+=2, printf("| "));
  printf("|-TRACE: %s()\n",__FUNCTION__);
#endif

  args[1] = iface;

  /* execute command */
  error = toolRunProgram(args);

  return error;
}

int ipv6IfaceStart(char *iface)
{
  int error;
  char *args[3]={IPV6_IFUP_CMD, NULL, NULL};

#ifdef TRACE
  DEBUG_indent += 2;
  for(DEBUG_i=0; DEBUG_i<(DEBUG_indent-2); DEBUG_i+=2, printf("| "));
  printf("|-TRACE: %s()\n",__FUNCTION__);
#endif

  args[1] = iface;

  /* execute command */
  error = toolRunProgram(args);

  return error;
}

int ipv6IfaceSetMac(char *iface, char *mac)
{
  int error;
  char *args[4]={IPV6_IFSETMAC_CMD, NULL, NULL, NULL};

#ifdef TRACE
  DEBUG_indent += 2;
  for(DEBUG_i=0; DEBUG_i<(DEBUG_indent-2); DEBUG_i+=2, printf("| "));
  printf("|-TRACE: %s()\n",__FUNCTION__);
#endif

  args[1] = iface;

  if(mac == NULL)
    {
      args[0] = IPV6_IFRESETMAC_CMD;
    }
  else
    {
      args[2] = mac;
    }
  
  /* execute command */
  error = toolRunProgram(args);

  return error;
}

int ipv6IfaceSetMtu(char *iface, char *mtu)
{
  int error;
  char *args[5]={IPV6_IFCONFIG_CMD, NULL, "mtu", NULL, NULL};

#ifdef TRACE
  DEBUG_indent += 2;
  for(DEBUG_i=0; DEBUG_i<(DEBUG_indent-2); DEBUG_i+=2, printf("| "));
  printf("|-TRACE: %s()\n",__FUNCTION__);
#endif

  args[1] = iface;

  if(mtu == NULL)
    {
      args[3] = IPV6_DEFAULT_MTU;
    }
  else
    {
      args[3] = mtu;
    }
  
  /* execute command */
  error = toolRunProgram(args);

  return error;
}

int ipv6IfaceSetAddr(char *iface, char *addr)
{
  int error;
  char *args[6]={IPV6_IFCONFIG_CMD, NULL, "inet6", "add", NULL, NULL};

#ifdef TRACE
  DEBUG_indent += 2;
  for(DEBUG_i=0; DEBUG_i<(DEBUG_indent-2); DEBUG_i+=2, printf("| "));
  printf("|-TRACE: %s()\n",__FUNCTION__);
#endif

#ifdef __linux__
  /* On Linux, network interface needs to be up before
     being assigned an IPv6 address */
  ipv6IfaceStart(iface);
#endif /* __linux__ */

  args[1] = iface;
  args[4] = addr;
  
  /* execute command */
  error = toolRunProgram(args);

  /* wait for the interface to setup */
  sleep(10);

  return error;
}

int ipv6NodeEnableRouter()
{
  int error;
  char *args[3]={IPV6_ROUTER_CMD, "on", NULL};

#ifdef TRACE
  DEBUG_indent += 2;
  for(DEBUG_i=0; DEBUG_i<(DEBUG_indent-2); DEBUG_i+=2, printf("| "));
  printf("|-TRACE: %s()\n",__FUNCTION__);
#endif

  /* execute command */
  error = toolRunProgram(args);

  return error;
}

int ipv6NodeDisableRouter()
{
  int error;
  char *args[3]={IPV6_ROUTER_CMD, "off", NULL};

#ifdef TRACE
  DEBUG_indent += 2;
  for(DEBUG_i=0; DEBUG_i<(DEBUG_indent-2); DEBUG_i+=2, printf("| "));
  printf("|-TRACE: %s()\n",__FUNCTION__);
#endif

  /* execute command */
  error = toolRunProgram(args);

  return error;
}




int ipv6IfaceSetRoute(char *dest, char *next)
{
  int error;
  char *args[6]={IPV6_ROUTE_CMD, "add", "-inet6", NULL, NULL, NULL};

#ifdef TRACE
  DEBUG_indent += 2;
  for(DEBUG_i=0; DEBUG_i<(DEBUG_indent-2); DEBUG_i+=2, printf("| "));
  printf("|-TRACE: %s()\n",__FUNCTION__);
#endif

  args[3] = dest;
  args[4] = next;
  
  /* execute command */
  error = toolRunProgram(args);

  return error;
}

int ipv6FlushRoutes()
{
  int error;
  char *args[4]={IPV6_ROUTE_CMD, "flush", "-inet6", NULL};

#ifdef TRACE
  DEBUG_indent += 2;
  for(DEBUG_i=0; DEBUG_i<(DEBUG_indent-2); DEBUG_i+=2, printf("| "));
  printf("|-TRACE: %s()\n",__FUNCTION__);
#endif

  /* execute command */
  error = toolRunProgram(args);

  return error;
}


#ifdef DEBUG
/**
 * \fn      void DEBUGprintpacket(struct ip6_hdr header, unsigned char *payload)
 * \brief   Print header informations of an IPv6 packet
 * \warning DEBUG PURPOSE ONLY !
 */
void DEBUGprintpacket(struct ip6_hdr header, unsigned char *payload, char *tag)
{
  int tmp, tmp2;
  int j;

  /* DEBUG PURPOSE ONLY */
  tmp = header.ip6_vfc >> 4;
  printf("%s VERSION:\t%d\n", tag, tmp);

  /* DEBUG PURPOSE ONLY */
  tmp = ntohl(header.ip6_flow & IPV6_TCLASS_MASK);
  tmp >>= 20;
  printf("%s TRAFFIC CLASS:\t%d (%X)\n", tag, tmp, tmp);

  /* DEBUG PURPOSE ONLY */
  tmp = ntohl(header.ip6_flow  & IPV6_FLOWLABEL_MASK);
  printf("%s FLOW LABEL:\t%d (%X)\n", tag, tmp, tmp);

  /* DEBUG PURPOSE ONLY */
  tmp2 = ntohs(header.ip6_plen);
  printf("%s PAYLOAD LENGTH:\t%d (%X)\n", tag, tmp2, tmp2);
  printf("%s NEXT HEADER: \t%d (%X)\n", tag, header.ip6_nxt, header.ip6_nxt);
  printf("%s HOP LIMIT: \t%d (%X)\n", tag, header.ip6_hlim, header.ip6_hlim);

  /* DEBUG PURPOSE ONLY */
  printf("%s SOURCE ADDRESS:\t", tag);
  for(j=0; j<14; j+=2)
    printf("%02X%02X:", header.ip6_src.s6_addr[j], header.ip6_src.s6_addr[j+1]);
  printf("%02X%02X", header.ip6_src.s6_addr[j], header.ip6_src.s6_addr[j+1]);
  printf("\n");

  /* DEBUG PURPOSE ONLY */
  printf("%s DEST. ADDRESS:\t", tag);
  for(j=0; j<14; j+=2)
    printf("%02X%02X:", header.ip6_dst.s6_addr[j], header.ip6_dst.s6_addr[j+1]);
  printf("%02X%02X", header.ip6_dst.s6_addr[j], header.ip6_dst.s6_addr[j+1]);
  printf("\n");

  /* DEBUG PURPOSE ONLY */
  printf("%s PAYLOAD:\t\t", tag);

  /* DEBUG PURPOSE ONLY */
  for(tmp=0; tmp<tmp2; tmp++)
    printf("%02X", (unsigned char)payload[tmp]);
  printf("\n");

}
#endif /* DEBUG */

int ipv6SendRaw(struct ip6_hdr *header, int nb_hop, struct in6_addr *hops, unsigned char *data)
{
  int error;

  struct ifaddrs  *ifa, *ifap;
  struct in6_addr *src_addr;

  struct msghdr msg;
  struct cmsghdr *cmsg=NULL;
  struct iovec send_iovec;
  struct sockaddr_in6 sock_addr;
  void *ptr=NULL;

#ifdef TRACE
  DEBUG_indent += 2;
  for(DEBUG_i=0; DEBUG_i<(DEBUG_indent-2); DEBUG_i+=2, printf("| "));
  printf("|-TRACE: %s()\n",__FUNCTION__);
#endif

  if(ipv6_socket != 0)
    close(ipv6_socket);

  /* get configured IPv6 addresses */
  error = getifaddrs(&ifap);
  if(error != 0)
    {
      warnx("%s() %d: %s", __FUNCTION__, __LINE__, strerror(errno));
      return IPV6_EXTERNAL_ERRCODE;
    }

  /* find the configured IPv6 address which match src address */
  for(ifa=ifap; ifa; ifa=ifa->ifa_next)
     {
      src_addr = &((struct sockaddr_in6 *)ifa->ifa_addr)->sin6_addr;

      /*if(memcmp(src_addr, &(header->ip6_src), sizeof(struct in6_addr)) == 0) */
      /* FreeBSD link-local addresses are recorded in a strange way */
      if((memcmp(src_addr, &(header->ip6_src), 2) == 0)
	 && (memcmp(((unsigned char *)src_addr) + 4, ((unsigned char *)&(header->ip6_src)) +4,
		    sizeof(struct in6_addr) - 4) == 0))
	break;
    }

  if(ifa == NULL)
    {
      warnx("%s(): %s", __FUNCTION__, IPV6_SEND_INVALIDSRC_ERRMSG); 
      return IPV6_SEND_INVALIDSRC_ERRCODE;
    }

  /* sniff the sent packet */
/*   dirty_snif(ifa->ifa_name, src_addr); */


  if(nb_hop != 0)
    {
      int i;

      ptr = malloc(inet6_rthdr_space(IPV6_RTHDR_TYPE_0, nb_hop));
      if(ptr == NULL)
	{
	  warnx("%s() %d: %s", __FUNCTION__, __LINE__, strerror(errno));
	  return IPV6_EXTERNAL_ERRCODE;
	}

      cmsg = inet6_rthdr_init(ptr, IPV6_RTHDR_TYPE_0);

      /* add initial destination to hop list */
      inet6_rthdr_add(cmsg, &header->ip6_dst, IPV6_RTHDR_LOOSE);

      /* add each initial hop to hop list except the last one */
      /* which is the final destination */
      for(i=0; i<(nb_hop-1); i++)
	{
	  inet6_rthdr_add(cmsg, &(hops[i]), IPV6_RTHDR_LOOSE);
	}
      inet6_rthdr_lasthop(cmsg, IPV6_RTHDR_LOOSE);

      /* set the final destination (last hop of the initial packet) */
      memcpy(&header->ip6_dst, &hops[i], sizeof(header->ip6_dst));

    }

  /* create an IPv6 raw socket */
  ipv6_socket = socket(PF_INET6, SOCK_RAW, header->ip6_nxt);
  if(ipv6_socket < 0)
    {
      warnx("%s() %d: %s", __FUNCTION__, __LINE__, strerror(errno));
      freeifaddrs(ifap);
      free(ptr);
      return IPV6_EXTERNAL_ERRCODE;
    }

  /* set some receive options */
  error = 1;
  setsockopt(ipv6_socket, IPPROTO_IPV6, IPV6_RECVPKTINFO, &error, sizeof(error));
  setsockopt(ipv6_socket, IPPROTO_IPV6, IPV6_RECVHOPLIMIT, &error, sizeof(error));
  setsockopt(ipv6_socket, IPPROTO_IPV6, IPV6_RECVTCLASS, &error, sizeof(error));

  setsockopt(ipv6_socket, IPPROTO_IPV6, IPV6_HOPOPTS, &error, sizeof(error));
  setsockopt(ipv6_socket, IPPROTO_IPV6, IPV6_DSTOPTS, &error, sizeof(error));
  setsockopt(ipv6_socket, IPPROTO_IPV6, IPV6_RTHDR, &error, sizeof(error));

  /* disable automatic flowlabel handling */
  error = 0;
  setsockopt(ipv6_socket, IPPROTO_IPV6, IPV6_AUTOFLOWLABEL, &error, sizeof(error));

  /* set Traffic Class value */
  error = ntohl(header->ip6_flow & IPV6_TCLASS_MASK);
  error >>= 20;

  setsockopt(ipv6_socket, IPPROTO_IPV6, IPV6_TCLASS, &error, sizeof(error));


  /* set Hop Limit value */
  setsockopt(ipv6_socket, IPPROTO_IPV6, IPV6_UNICAST_HOPS,
	     &header->ip6_hlim, sizeof(header->ip6_hlim));


  /* bind the previously created socket to an IPv6 address */
  error = bind(ipv6_socket, ifa->ifa_addr, sizeof(struct sockaddr_in6));
  if(error != 0)
    {
      warnx("%s() %d: %s", __FUNCTION__, __LINE__, strerror(errno));
      freeifaddrs(ifap);
      free(ptr);
      return IPV6_EXTERNAL_ERRCODE;
    }

  freeifaddrs(ifap);

  /* destination address */
  bzero(&sock_addr, sizeof(struct sockaddr_in6));
    sock_addr.sin6_len = sizeof(struct sockaddr_in6);
  sock_addr.sin6_family = AF_INET6;
  sock_addr.sin6_flowinfo = 0xABCDEFAD;
  sock_addr.sin6_addr = header->ip6_dst;

  /* send IPv6 packet */
  send_iovec.iov_base = data;  
  send_iovec.iov_len  = ntohs(header->ip6_plen);

  msg.msg_name    = &sock_addr;
  msg.msg_namelen = sizeof(struct sockaddr_in6);
  msg.msg_iov     = &send_iovec;
  msg.msg_iovlen  = 1;
  msg.msg_control = ptr;

  if(cmsg==NULL)
    {
      msg.msg_controllen = 0;
    }
  else
    {
      msg.msg_controllen = cmsg->cmsg_len;
    }

  error = sendmsg(ipv6_socket, &msg, 0);
  if(error < 0)
    {
      warnx("%s() %d: %s", __FUNCTION__, __LINE__, strerror(errno));
      free(ptr);
      return IPV6_EXTERNAL_ERRCODE;
    }

  free(ptr);

  return 0;
}

int ipv6Decode(unsigned char *data)
{
  int i;
  int next;
  int nb_hop=0;
  int result;

  struct ip6_hdr header;
  struct ip6_hbh hbh_header;
  struct ip6_dest dest_header;
  struct ip6_rthdr0 rt_header;
  struct in6_addr *hops=NULL;
  
#ifdef TRACE
  DEBUG_indent += 2;
  for(DEBUG_i=0; DEBUG_i<(DEBUG_indent-2); DEBUG_i+=2, printf("| "));
  printf("|-TRACE: %s()\n",__FUNCTION__);
#endif

  memcpy(&header, data, sizeof(header));
  
  /* skip header */
  data = &data[sizeof(header)];

#ifdef DEBUG
  DEBUGprintpacket(header, data, "DEC ");
#endif


  /* check IP version */
  if((header.ip6_vfc & IPV6_VERSION_MASK) != IPV6_VERSION) 
    {
      warnx("%s(): %s", __FUNCTION__, IPV6_SEND_INVALIDVERSION_ERRMSG);
      return IPV6_SEND_INVALIDVERSION_ERRCODE;
    }

  /* check extension headers */
  next = header.ip6_nxt;
  while(next >= 0)
    {
      switch(next)
	{
	case IPPROTO_HOPOPTS:
	  /* At the time being, all hop by hop options are */
	  /* managed by the kernel... */
	  
	  /* ignore this header */
	  memcpy(&hbh_header, data, sizeof(hbh_header));
	  
	  /* skip header */
	  data = &data[(8 * hbh_header.ip6h_len) + sizeof(hbh_header) ]; 
	  header.ip6_plen = htons(ntohs(header.ip6_plen) - ((8 * hbh_header.ip6h_len) + sizeof(hbh_header)));
	  
	  /* update next value */
	  next = hbh_header.ip6h_nxt;
	  
	  break;
	  
	case IPPROTO_DSTOPTS:
	  /* At the time being, no destination options are */
	  /* available */
	  
	  /* ignore this header */
	  memcpy(&dest_header, data, sizeof(dest_header));
	  
	  /* skip header */
	  data = &data[(8 * dest_header.ip6d_len) + sizeof(dest_header) ]; 
	  header.ip6_plen = htons(ntohs(header.ip6_plen) - ((8 * dest_header.ip6d_len) + sizeof(dest_header)));
	  
	  /* update next value */
	  next = dest_header.ip6d_nxt;
	  
	  break;

	case IPPROTO_ROUTING:

	  /* TODO : sizeof(rt_header) is different in FBSD and Linux */
	  memcpy(&rt_header, data, 8/*sizeof(rt_header)*/);
	  
	  /* skip static part of this header */
	  data = &data[8 /*sizeof(rt_header)*/]; 

	  header.ip6_plen = htons(ntohs(header.ip6_plen) - 8/*sizeof(rt_header)*/); 
	  
	  /* update next value */
	  next = rt_header.ip6r0_nxt;
	  
	  /* store hops list */
	  nb_hop = (rt_header.ip6r0_len/2);
	  hops = (struct in6_addr *)malloc(nb_hop*sizeof(struct in6_addr));
	  if(hops == NULL)
	    {
	      warnx("%s(): %s", __FUNCTION__, strerror(errno));
	      return IPV6_EXTERNAL_ERRCODE;
	    }

	  for(i=0; i<nb_hop; i++)
	    {
	      memcpy(&hops[i], data, sizeof(struct in6_addr));
	      data = &data[sizeof(struct in6_addr)];
	      header.ip6_plen = htons(ntohs(header.ip6_plen) - sizeof(struct in6_addr));
	    }
	  break;

	case IPPROTO_AH:
	case IPPROTO_ESP:
	case IPPROTO_FRAGMENT:
	  /* No information about fragment header */

	  /* TODO */

	  break;
	  
	default:
	  next = 0 - next;
	  break;
	}
    }
     
  header.ip6_nxt = 0 - next;

  /* check payload */
  switch(header.ip6_nxt)
    {
    case IPPROTO_ICMPV6:
      /* check if it is an "Echo Request" */
      if( (unsigned char)(*data) == ICMP6_ECHO_REQUEST)
	{
	  result = ipv6SendPing(&header, nb_hop, hops);
	  /* result = ipv6SendRaw(&header, nb_hop, hops, data); */
	}
      else
	{
	  result = ipv6SendRaw(&header, nb_hop, hops, data);
	}
      break;

    default:
      result = ipv6SendRaw(&header, nb_hop, hops, data);
      break;
    }
  
  free(hops);

  return result;
}

int ipv6Get(char **result)
{
  int len;
  int i;
  struct msghdr msg;
  struct iovec recv_iovec;
  struct sockaddr_in6 sock_addr;
  unsigned char buf[1024], packet[5000];
  struct ip6_hdr header;
  
  struct cmsghdr *cmsg=NULL;
  
#ifdef TRACE
  DEBUG_indent += 2;
  for(DEBUG_i=0; DEBUG_i<(DEBUG_indent-2); DEBUG_i+=2, printf("| "));
  printf("|-TRACE: %s()\n",__FUNCTION__);
#endif

  memset(&sock_addr, 0, sizeof(sock_addr));

  msg.msg_name = (caddr_t)&sock_addr;
  msg.msg_namelen = sizeof(sock_addr);

  memset(&recv_iovec, 0, sizeof(recv_iovec));
  recv_iovec.iov_base = (caddr_t)packet;
  recv_iovec.iov_len = sizeof(packet);

  msg.msg_iov = &recv_iovec;
  msg.msg_iovlen = 1;

  msg.msg_control = (caddr_t)buf;
  msg.msg_controllen = sizeof(buf);

  len = recvmsg(ipv6_socket, &msg, 0);
  
  /* retrieve paylaod length */
  header.ip6_plen = htons(len);

  /* set ip version */
  header.ip6_vfc = 0x60;

  /* retrieve hoplimit */
  for(cmsg = (struct cmsghdr *)CMSG_FIRSTHDR(&msg); cmsg; cmsg = (struct cmsghdr *)CMSG_NXTHDR(&msg, cmsg))
    {
      if(cmsg->cmsg_level == IPPROTO_IPV6 &&
	 cmsg->cmsg_type == IPV6_HOPLIMIT &&
	 cmsg->cmsg_len == CMSG_LEN(sizeof(int)))
	header.ip6_hlim = (int)*CMSG_DATA(cmsg);
    }
  
  /* retrieve traffic class */
  for(cmsg = (struct cmsghdr *)CMSG_FIRSTHDR(&msg); cmsg; cmsg = (struct cmsghdr *)CMSG_NXTHDR(&msg, cmsg))
    {
      if(cmsg->cmsg_level == IPPROTO_IPV6 &&
	 cmsg->cmsg_type == IPV6_TCLASS &&
	 cmsg->cmsg_len == CMSG_LEN(sizeof(int)))
	{
	  i = (int)*CMSG_DATA(cmsg);
	  i <<= 16;
	  header.ip6_flow |= i;
	}
    }

  /* retrieve destination address */
  for(cmsg = (struct cmsghdr *)CMSG_FIRSTHDR(&msg); cmsg; cmsg = (struct cmsghdr *)CMSG_NXTHDR(&msg, cmsg)) 
    {
      if(cmsg->cmsg_level == IPPROTO_IPV6 &&
	 cmsg->cmsg_type == IPV6_PKTINFO && 
	 cmsg->cmsg_len == CMSG_LEN(sizeof(struct in6_pktinfo)))
	memcpy(&(header.ip6_dst.s6_addr), &(((struct in6_pktinfo *)CMSG_DATA(cmsg))->ipi6_addr), sizeof(struct in6_addr));
    }

  /* retrieve source address */
  memcpy(&(header.ip6_src.s6_addr), &sock_addr.sin6_addr, sizeof(struct in6_addr));

  /* store result */
  (*result) = (char *)malloc( 2 * (sizeof(struct ip6_hdr) + len) + 1);
  for(i=0; i<sizeof(struct ip6_hdr); i++)
    sprintf( &(*result)[i*2], "%02X", ((unsigned char *)&header)[i]);

  for(i=0; i<len; i++)
    sprintf( &(*result)[(sizeof(struct ip6_hdr) + i)*2], "%02X", ((unsigned char *)&packet)[i]);

  (*result)[2 * (sizeof(struct ip6_hdr) + len)]='\0';

#ifdef DEBUG
  DEBUGprintpacket(header, packet, "RECV");
#endif /* DEBUG */

  return 0;
}
