/**
 * \file
 * \brief SCOP Decode module (SCOP Interface)
 */


#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <err.h>
#include <errno.h>

#include <math.h>

#include "scop.h"


int scopDecode()
{
  int result;
  char *msg;

#ifdef TRACE
  DEBUG_indent += 2;
  for(DEBUG_i=0; DEBUG_i<(DEBUG_indent-2); DEBUG_i+=2, printf("| "));
  printf("|-TRACE: %s()\n",__FUNCTION__);
#endif

  /* extract SCOP headers */
  scopExtractHeader(&msg);

  /* check if message begins with the SCOP magic string */
  if(strncmp(msg, SCOP_MAGIC, strlen(SCOP_MAGIC)) != 0)
    {
      warnx("%s(): %s", __FUNCTION__, SCOP_DECODE_NOTSCOP_ERRMSG);
      global_errmsg = SCOP_DECODE_NOTSCOP_ERRMSG;
      return SCOP_DECODE_NOTSCOP_ERRCODE;
    }

  /* check if message SCOP version is equal to our SCOP version */
  if(strncmp(msg+strlen(SCOP_MAGIC), SCOP_VERSION, strlen(SCOP_VERSION)) != 0)
    {
      warnx("%s(): %s", __FUNCTION__, SCOP_DECODE_BADSCOPVERSION_ERRMSG);
      global_errmsg = SCOP_DECODE_BADSCOPVERSION_ERRMSG;
      return SCOP_DECODE_BADSCOPVERSION_ERRCODE;
    }

  /* msg seems to be a valid SCOP message */
  free(msg);
  
  /* extract message type */
  result = scopExtractShortString(&msg);
  if(result != 0)
    {
      return result;
    }

  /* check if message is an InitRequest */
  if(strncmp(msg, SCOP_INIT_IUT_RQ, strlen(SCOP_INIT_IUT_RQ)) == 0)
    {
      free(msg);
      result = scopDecodeInitIutRequest();
      return result;
    }

  /* check if message is a ResetRequest */
  if(strncmp(msg, SCOP_RESET_RQ, strlen(SCOP_RESET_RQ)) == 0)
    {
      free(msg);
      result = scopDecodeInitIutRequest();
      return result;
    }

  /* check if message is a DataRequest */
  if(strncmp(msg, SCOP_DATA_RQ, strlen(SCOP_DATA_RQ)) == 0)
    {
      free(msg);
      result = scopDecodeDataRequest();
      return result;
    }

  /* Ooops : unknown message type */
  warnx("%s(): %s", __FUNCTION__, SCOP_DECODE_BADTYPE_ERRMSG);
  free(msg);

  global_errmsg = SCOP_DECODE_BADTYPE_ERRMSG;
  return SCOP_DECODE_BADTYPE_ERRCODE;
  
}


int scopDecodeInitIutRequest()
{
  char *param;
  int result;
  char * tc_id = NULL;

#ifdef TRACE
  DEBUG_indent += 2;
  for(DEBUG_i=0; DEBUG_i<(DEBUG_indent-2); DEBUG_i+=2, printf("| "));
  printf("|-TRACE: %s()\n",__FUNCTION__);
#endif

  /* extract each param */
  while(scopExtractShortString(&param) == 0)
    {
      if(strncmp(param, "testCaseId", 10) == 0)
	{
	  /* check if this param has been previously received */
	  if(tc_id != NULL)
	    {
	      warnx("%s(): %s", __FUNCTION__, SCOP_DECODE_PARAMREPEAT_ERRMSG);
	      global_errmsg = SCOP_DECODE_PARAMREPEAT_ERRMSG;
	      return SCOP_DECODE_PARAMREPEAT_ERRCODE;
	    }
	  
	  /* extract tcId value */
	  result = scopExtractShortString(&tc_id);
	  if(result != 0)
	    {
	      free(param);
	      return result;
	    }
	  
#ifdef DEBUG
	  printf("DEBUG: %s() -> %s: %s\n",__FUNCTION__, param, tc_id);
#endif

	  free(param);
	  continue;
	}

      if(strncmp(param, "params", 6) == 0)
	{
	  /* scopExtractList(&msg, &msg_len, &value); */
	  
	  /* TODO */

	  free(param);
	  continue;
	}

      /* Ooops : unknown parameter */
      warnx("%s(): %s '%s'", __FUNCTION__, SCOP_DECODE_BADPARAM_ERRMSG, param);
      free(param);

      global_errmsg = SCOP_DECODE_BADPARAM_ERRMSG;
      return SCOP_DECODE_BADPARAM_ERRCODE;
    }

  /* check if this all params have been received */
  if(tc_id == NULL)
    {
      warnx("%s(): %s", __FUNCTION__, SCOP_DECODE_MISSINGPARAM_ERRMSG);
      global_errmsg = SCOP_DECODE_MISSINGPARAM_ERRMSG;
      return SCOP_DECODE_MISSINGPARAM_ERRCODE;
    }

  /* Initialize IUT */
  result = scopInitRequest(tc_id);

  free(tc_id);
  return result;
}

int scopDecodeResetRequest()
{
  int result;
#ifdef TRACE
  DEBUG_indent += 2;
  for(DEBUG_i=0; DEBUG_i<(DEBUG_indent-2); DEBUG_i+=2, printf("| "));
  printf("|-TRACE: %s()\n",__FUNCTION__);
#endif

  /* reset IUT */
  result = scopResetRequest();
  return result;
}


int scopDecodeDataRequest()
{
  char *param, *tmp;
  int result, i;

  char * cmd = NULL;
  unsigned char * data = NULL;


#ifdef TRACE
  DEBUG_indent += 2;
  for(DEBUG_i=0; DEBUG_i<(DEBUG_indent-2); DEBUG_i+=2, printf("| "));
  printf("|-TRACE: %s()\n",__FUNCTION__);
#endif

  /* extract each param */
  while(scopExtractShortString(&param) == 0)
    {
      if(strncmp(param, "command", 7) == 0)
	{
	  /* check if this param has been previously received */
	  if(cmd != NULL)
	    {
	      warnx("%s(): %s", __FUNCTION__, SCOP_DECODE_PARAMREPEAT_ERRMSG);
	      global_errmsg = SCOP_DECODE_PARAMREPEAT_ERRMSG;
	      return SCOP_DECODE_PARAMREPEAT_ERRCODE;
	    }

	  /* extract cmd value */
	  result = scopExtractShortString(&cmd);
	  if(result != 0)
	    {
	      free(param);
	      return result;
	    }
	  
#ifdef DEBUG
	  printf("DEBUG: %s() -> %s: %s\n",__FUNCTION__, param, cmd);
#endif

	  free(param);

	  continue;
	}

      if(strncmp(param, "data", 4) == 0)
	{
	  /* check if this param has been previously received */
	  if(data != NULL)
	    {
	      warnx("%s(): %s", __FUNCTION__, SCOP_DECODE_PARAMREPEAT_ERRMSG);
	      global_errmsg = SCOP_DECODE_PARAMREPEAT_ERRMSG;
	      return SCOP_DECODE_PARAMREPEAT_ERRCODE;
	    }

	  /* extract data value */
	  result = scopExtractLongString(&tmp);
	  if(result != 0)
	    {
	      free(param);
	      return result;
	    }
	  
#ifdef DEBUG
	  printf("DEBUG: %s() -> %s: %s\n",__FUNCTION__, param, tmp);
#endif
	  
	  /* convert ascii data into real octet string */
	  data = (unsigned char *)malloc(strlen(tmp)/2);
	  if(data==NULL)
	    {
	      free(param);
	      free(tmp);
	      warnx("%s(): %s", __FUNCTION__, strerror(errno));
	      global_errmsg = strerror(errno);
	      return SCOP_EXTERNAL_ERRCODE;
	    }
	  for(i=(strlen(tmp)/2)-1; i>=0 ; i--)
	    {
	      /* TODO : check errors here ! */
	      data[i] = (unsigned char)strtol(&tmp[i*2], NULL, 16);
	      tmp[i*2] = '\0';
	    }

	  free(tmp);
	  free(param);
	  continue;
	}


      if(strncmp(param, "params", 6) == 0)
	{
	  /* scopExtractList(&msg, &msg_len, &value); */

	  /* TODO */

	  free(param);
	  continue;
	}

      /* Ooops : unknown parameter */
      warnx("%s(): %s %s", __FUNCTION__, SCOP_DECODE_BADPARAM_ERRMSG, param);
      free(param);
      
      global_errmsg = SCOP_DECODE_BADPARAM_ERRMSG;
      return SCOP_DECODE_BADPARAM_ERRCODE;

    }

  /* check if this all params have been received */
  if((cmd == NULL))
    {
      warnx("%s(): %s", __FUNCTION__, SCOP_DECODE_MISSINGPARAM_ERRMSG);
      global_errmsg = SCOP_DECODE_MISSINGPARAM_ERRMSG;
      return SCOP_DECODE_MISSINGPARAM_ERRCODE;
    }

  /* Execute DataRequest */
  result = scopDataRequest(cmd, data);

  free(cmd);
  free(data);

  return(result);
}

int scopExtractHeader(char **string)
{
  unsigned int header_len;

#ifdef TRACE
  DEBUG_indent += 2;
  for(DEBUG_i=0; DEBUG_i<(DEBUG_indent-2); DEBUG_i+=2, printf("| "));
  printf("|-TRACE: %s()\n",__FUNCTION__);
#endif

  /* synchronize Transport layer using SCOP_MAGIC */
  scopTransportSynchro(SCOP_MAGIC);

  /* compute the expected header length */
  header_len = strlen(SCOP_MAGIC) + strlen(SCOP_VERSION);

  /* alloc memory to store header */
  (*string) = (char*)malloc(header_len + 1);
  if((*string) == NULL)
    {
      warnx("%s(): %s", __FUNCTION__, strerror(errno));
      global_errmsg = strerror(errno);
      return SCOP_EXTERNAL_ERRCODE;
    }

  /* get header from transport module */
  strncpy(*string, SCOP_MAGIC, strlen(SCOP_MAGIC));
  scopTransportRead(&((*string)[strlen(SCOP_MAGIC)]), strlen(SCOP_VERSION));

  (*string)[header_len] = '\0';
  
  return 0;
}

int scopExtractShortString(char **string)
{
  unsigned int string_len;

#ifdef TRACE
  DEBUG_indent += 2;
  for(DEBUG_i=0; DEBUG_i<(DEBUG_indent-2); DEBUG_i+=2, printf("| "));
  printf("|-TRACE: %s()\n",__FUNCTION__);
#endif

  /* extract string length */
  string_len = 0;
  scopTransportRead((char *)&string_len, 1);

  if(string_len == 0)
    {
      (*string) = NULL;
      return SCOP_DECODE_EOM;
    }

  /* extract string */
  (*string) = (char*)malloc(string_len + 1);
  if((*string) == NULL)
    {
      warnx("%s(): %s", __FUNCTION__, strerror(errno));
      global_errmsg = strerror(errno);
      return SCOP_EXTERNAL_ERRCODE;
    }
  
  scopTransportRead((*string), string_len);
  (*string)[string_len]='\0';

  return 0;
}

int scopExtractLongString(char **string)
{
  int tmp, string_len;

#ifdef TRACE
  DEBUG_indent += 2;
  for(DEBUG_i=0; DEBUG_i<(DEBUG_indent-2); DEBUG_i+=2, printf("| "));
  printf("|-TRACE: %s()\n",__FUNCTION__);
#endif

  /* extract string length */
  tmp = 0;
  scopTransportRead((char *)&tmp, 1);
  string_len = tmp * pow(256, 2);

  tmp = 0;
  scopTransportRead((char *)&tmp, 1);
  string_len += tmp * pow(256, 1);

  tmp = 0;
  scopTransportRead((char *)&tmp, 1);
  string_len += tmp * pow(256, 0);

  tmp = 0;
  scopTransportRead((char *)&tmp, 1);
  string_len += tmp;

  /* extract string */
  (*string) = (char*)malloc(string_len + 1);
  if((*string) == NULL)
    {
      warnx("%s(): %s", __FUNCTION__, strerror(errno));
      global_errmsg = strerror(errno);
      return SCOP_EXTERNAL_ERRCODE;
    }
  
  scopTransportRead((*string), string_len);
  (*string)[string_len]='\0';

  return 0;
}

/* TODO check memory */
/* TODO store parameters */
int scopExtractList()
{

#ifdef TRACE
  DEBUG_indent += 2;
  for(DEBUG_i=0; DEBUG_i<(DEBUG_indent-2); DEBUG_i+=2, printf("| "));
  printf("|-TRACE: %s()\n",__FUNCTION__);
#endif

  /* TODO */



  return 0;
}







