/**
 * \file
 * \brief Tools module (Core Component)
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <err.h>
#include <errno.h>

#include <fcntl.h>

#include <sys/types.h>
#include <sys/wait.h>

#include "tools.h"


int toolRunProgram(char **args)
{
  pid_t pid;
  int status;

  /* create a new processus */
  pid = fork();
  if(pid == -1)
    {
      warnx("%s(): %s", __FUNCTION__, strerror(errno));
      return TOOL_EXTERNAL_ERRCODE;
    }
  if(pid == 0)
    {
      /* in the new processus, execute external command */ 
      execv(args[0], args);
      warnx("%s(): %s", __FUNCTION__, strerror(errno));
      exit(TOOL_EXTERNAL_ERRCODE);
    }
  
  /* wait for the new processus to exit */
  if(waitpid(pid, &status, 0) == -1)
    {
      warnx("%s(): %s", __FUNCTION__, strerror(errno));
      return TOOL_EXTERNAL_ERRCODE;
    }

  /* check exit status */
  if(WIFEXITED(status))
    {
      if(WEXITSTATUS(status) != 0)
	{
	  warnx("%s(): %s %d", __FUNCTION__, "Child exited with code ", WEXITSTATUS(status));
	  return TOOL_EXTERNAL_ERRCODE;      
	}
      return 0;
    }

  /* FIXME */
  return 0;
  /*return TOOL_UNKNOWN_ERRCODE;*/
}

int toolLoadFile(int fd, char ** buf)
{
  int nb;
  off_t pos;

/* #ifdef DEBUG */
/*   printf("DEBUG: %s()\n",__FUNCTION__); */
/* #endif */

  pos = lseek(fd, 0, SEEK_END);
  if(pos == -1)
    {
      warnx("%s(): %s", __FUNCTION__, strerror(errno));
      return TOOL_EXTERNAL_ERRCODE;
    }

  nb = (int)pos;

  (*buf) = (char *)malloc(nb + 1);
  if((*buf)==NULL)
    {
      warnx("%s(): %s", __FUNCTION__, strerror(errno));
      return TOOL_EXTERNAL_ERRCODE;
    }

  pos = lseek(fd, 0, SEEK_SET);
  if(pos == -1)
    {
      warnx("%s(): %s", __FUNCTION__, strerror(errno));
      return TOOL_EXTERNAL_ERRCODE;
    }

  nb = read(fd, (*buf), nb);
  if(nb == -1)
    {
      warnx("%s(): %s", __FUNCTION__, strerror(errno));
      return TOOL_EXTERNAL_ERRCODE;
    }

  (*buf)[nb] = '\0';
  
  return 0;
}

/* TODO : add some comments */
int toolGetTag(char *buf, char *tag, int nb, char ** result)
{
  int i, j, k, opened;
  int start=-1, end=-1;
  char *ftag=NULL;

/* #ifdef DEBUG */
/*   printf("DEBUG: %s()\n",__FUNCTION__); */
/* #endif */

  *result = NULL;

  while(start<0)
    {
      free(ftag);
      for(i=end+1; buf[i] && buf[i]!='<'; i++);
      if(buf[i]=='\0')
	return 1;

      if((strncmp(tag, &buf[i+1], strlen(tag))==0) && buf[i+1+strlen(tag)]=='>' && (--nb == 0))
	{
	  start = i + strlen(tag) + 2;
	  ftag = tag;
	}
      else
	{
	  for(j=++i; buf[j] && buf[j]!='>'; j++);
	  if(buf[j]=='\0')
	    return 1;

	  ftag = (char *)malloc(j - i + 1);
	  if(ftag==NULL)
	    {
	      warnx("%s(): %s", __FUNCTION__, strerror(errno));
	      return TOOL_EXTERNAL_ERRCODE;
	    }
	  for(k=i; k<j; k++)
	    ftag[k-i] = buf[k];
	  ftag[k-i] = '\0';
	}

      opened = 0;
      end = -1;
      while(opened!=0 || end<0)
	{
	  for(i++; buf[i] && buf[i]!='<'; i++);
	  if(buf[i+1]=='/')
	    {
	      if(opened==0)
		{
		  /* it should be our closing tag */
		  if((strncmp(ftag, &buf[i+2], strlen(ftag))==0) && buf[i+2+strlen(ftag)]=='>')
		    {
		      end = i;
		    }
		  else
		    {
		      /* parse error */
		      warnx("%s(): %s", __FUNCTION__, TOOL_XMLPARSE_ERRMSG);
		      return TOOL_XMLPARSE_ERRCODE;
		    }  
		}
	      else
		opened--;
	    }
	  else
	    opened++;
	}
    }

  (*result) = (char *)malloc(end - start + 1);
  if((*result)==NULL)
    {
      warnx("%s(): %s", __FUNCTION__, strerror(errno));
      return TOOL_EXTERNAL_ERRCODE;
    }
  for(i=start; i<end; i++)
    (*result)[i-start] = buf[i];
  (*result)[i-start] = '\0';

  return 0;

}


