/*
 * ----------------------------------------------------------------------------
 *  (C) Copyright Testing Technologies, 2001-2007.  All Rights Reserved.
 *
 *  All copies of this program, whether in whole or in part, and whether
 *  modified or not, must display this and all other embedded copyright
 *  and ownership notices in full.
 *
 *  See the file COPYRIGHT for details of redistribution and use.
 *
 *  You should have received a copy of the COPYRIGHT file along with
 *  this file; if not, write to the Testing Technologies,
 *  Rosenthaler Str. 13, 10119 Berlin, Germany.
 *
 *  TESTING TECHNOLOGIES DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
 *  SOFTWARE. IN NO EVENT SHALL TESTING TECHNOLOGIES BE LIABLE FOR ANY
 *  SPECIAL, DIRECT, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 *  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 *  AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 *  ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
 *  THIS SOFTWARE.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND,
 *  EITHER EXPRESSED OR IMPLIED, INCLUDING ANY KIND OF IMPLIED OR
 *  EXPRESSED WARRANTY OF NON-INFRINGEMENT OR THE IMPLIED WARRANTIES
 *  OF MECHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * ----------------------------------------------------------------------------
 *
 *  AUTHOR:      Dirk Tepelmann
 *  DATE:        January, 2005
 *
 *  REVISION INFO:
 *        $Revision: 1.6 $ $Date: 2005/09/13 08:47:39 $
 * ----------------------------------------------------------------------------- */
package com.testingtech.ttcn.tri.extFunc;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.math.BigInteger;
import java.net.Inet4Address;
import java.net.Inet6Address;
import java.net.InetAddress;
import java.net.UnknownHostException;

import org.etsi.ttcn.tci.CharstringValue;
import org.etsi.ttcn.tci.EnumeratedValue;
import org.etsi.ttcn.tci.IntegerValue;
import org.etsi.ttcn.tci.OctetstringValue;
import org.etsi.ttcn.tci.RecordOfValue;
import org.etsi.ttcn.tci.RecordValue;
import org.etsi.ttcn.tci.UnionValue;
import org.etsi.ttcn.tci.Value;
import org.etsi.ttcn.tri.TriMessage;

import com.testingtech.generated.codec.LibCommon_BasicTypesAndValues.TabularUInt8;
import com.testingtech.generated.codec.LibCommon_DataStrings.TabularOct1to255;
import com.testingtech.generated.codec.LibIpv6_Interface_TypesAndValues.TabularEspHeader;
import com.testingtech.generated.codec.LibIpv6_Interface_TypesAndValues.TabularEspIpDatagram;
import com.testingtech.generated.codec.LibIpv6_Interface_TypesAndValues.TabularEspPayload;
import com.testingtech.generated.codec.LibIpv6_Interface_TypesAndValues.TabularIpv6Header;
import com.testingtech.generated.codec.LibIpv6_Interface_TypesAndValues.TabularIpv6Payload;
import com.testingtech.generated.codec.LibIpv6_Interface_TypesAndValues.TabularMobileHeader;
import com.testingtech.generated.codec.LibIpv6_Rfc4306Ikev2_TypesAndValues.TabularEncryptedPayload;
import com.testingtech.generated.codec.LibIpv6_Rfc4306Ikev2_TypesAndValues.TabularIkePayloadList;
import com.testingtech.generated.codec.LibIpv6_Rfc4306Ikev2_TypesAndValues.TabularIkev2Header;
import com.testingtech.ttcn.logging.RTLoggingConstants;
import com.testingtech.ttcn.tci.codec.tabular.TabularEncoder;
import com.testingtech.ttcn.tci.codec.tabular.TabularException;
import com.testingtech.ttcn.tci.codec.tabular.TabularOctetstring;
import com.testingtech.ttcn.tri.ExternalFunctionsDefinition;
import com.testingtech.ttcn.tri.IPv6TestAdapter;
import com.testingtech.ttcn.tri.diffie_hellman.Primes;
import com.testingtech.ttcn.tri.encrypt.Encryption;
import com.testingtech.ttcn.tri.enums.IEnumDefinitions;
import com.testingtech.util.BitArrayOutputStream;

import de.tu_berlin.cs.uebb.muttcn.runtime.BigInt;
import de.tu_berlin.cs.uebb.muttcn.runtime.RB;
import de.tu_berlin.cs.uebb.ttcn.runtime.BitString;
import de.tu_berlin.cs.uebb.ttcn.runtime.OctetString;

//For debugging only
import com.testingtech.util.HexViewer;

/**
 * 
 * This class implements all the external function declaration of the
 * IPv6 ATS(STF276)
 *  
 */
public class ExternalFunctions_IPv6 implements ExternalFunctionsDefinition, IEnumDefinitions, Primes {
	
	private RB RB;
	
	private final int HOP = 0;
	private final int TUNNELEDIPV6 = 41;
	private final int ROUTING = 43;
	private final int FRAGMENT = 44;
	private final int DESTINATION = 60;
    
	private final int ipv6HeaderSize = 40;

	private static final String IKE_PRFKEY_LEN = "prfKeyLen";

	private static final String IKE_INTEGR_KEY_LEN = "ikeIntegrKeyLen";

	private static final String IKE_ENCR_KEY_LEN = "ikeEncrKeyLen";

	public ExternalFunctions_IPv6(RB rb)
	{
		RB = rb;
	}
	
	/**
	 * This external function translates an IPv6 address string in its octetstring representation.
	 * @param p_address
	 * @return
	 */
	public synchronized OctetstringValue fx_translateIpv4Address(CharstringValue p_address) {
	    InetAddress addr;
	    byte bmsg[] = new byte[]{};
	    String pAddr = p_address.getString();
	    
	    if (pAddr.equals("")) {
            RB.getTciTMProvided().tciError("fx_translateIpv4Address: Cannot translate an empty string for " + p_address.getType().getName());
	    }
	    
	    try {
	    	addr = Inet4Address.getByName(pAddr);
	    	bmsg = addr.getAddress();
	    } catch (UnknownHostException e) {
            RB.getTciTMProvided().tciError("fx_translateIpv4Address: Unknown Host: " + e.getMessage());
	    }
	    
		OctetString ret = (OctetString) RB.getTciCDRequired().getTypeForName("LibIpv6_CommonRfcs_TypesAndValues.Ipv4Address").newInstance();
	    if (bmsg.length != 0) {
	    	ret.setOctetstring(bmsg);
	    }
	    ret.setLength(4);
	    
	    return ret;		
	}

	/**
	 * This external function translates an IPv6 address string in its octetstring representation.
	 * @param p_address
	 * @return
	 */
	public synchronized OctetstringValue fx_translateIpv6Address(CharstringValue p_address) {
	    InetAddress addr;
	    byte bmsg[] = new byte[]{};
	    String pAddr = p_address.getString();
	    
	    if (pAddr.equals("")) {
            RB.getTciTMProvided().tciError("fx_translateIpv6Address: Cannot translate an empty string for " + p_address.getType().getName());
	    }
	    
	    try {
	    	addr = Inet6Address.getByName(pAddr);
	    	bmsg = addr.getAddress();
	    } catch (UnknownHostException e) {
            RB.getTciTMProvided().tciError("fx_translateIpv6Address: Unknown Host: " + e.getMessage());
	    }
	    
		OctetString ret = (OctetString) RB.getTciCDRequired().getTypeForName("LibIpv6_CommonRfcs_TypesAndValues.Ipv6Address").newInstance();
	    if (bmsg.length != 0) {
	    	ret.setOctetstring(bmsg);
	    }
	    ret.setLength(16);
	    
	    return ret;		
	}

	/**
	 * This external function calculates the payload length 
	 * of a IPv6 packet
	 * 
	 * @param   p_msg 	The IPv6 packet for calculation.
	 * @return  		The payload length.
	 */
	public synchronized IntegerValue fx_payloadLength(RecordValue p_msg ){

		try {
		    TriMessage message = RB.getTciEncoding().getCodec("LibIpv6_Interface_TypesAndValues").encode(p_msg);
		    
		    int msgLen = message.getEncodedMessage().length - ipv6HeaderSize;
		    
		    IntegerValue ret = (IntegerValue) RB.getTciCDRequired().getTypeForName("LibCommon_BasicTypesAndValues.UInt16").newInstance();

		    ret.setInt(msgLen);
		    
		    return ret;
		} catch (Exception e) {
			e.printStackTrace();
		}
		return null;
	}

	/**
	 * This external function calculates the payload length 
	 * of a IPv6 packet
	 * 
	 * @param   p_msg 	The IPv6 packet for calculation.
	 * @param   p_entry   The indicator for which tunnelled header
	 *                  the payload length should be calculated.
	 * @return  		The payload length.
	 */
	public synchronized IntegerValue fx_tunnelledPayloadLength(RecordValue p_msg, IntegerValue p_entry){
	    
		int v_cnt = 0;
		
	    TriMessage message = RB.getTciEncoding().getCodec("LibIpv6_Interface_TypesAndValues").encode(p_msg);
	    
	    int msgLen = message.getEncodedMessage().length - ipv6HeaderSize;
	    
	    TabularIpv6Header ipv6Header = new TabularIpv6Header(p_msg.getField("ipv6Hdr"));
	    try {
            ipv6Header.encode(new TabularEncoder(new BitArrayOutputStream(), RB));
        } catch (TabularException tex) {
            RB.getTciTMProvided().tciError(tex.getMessage());
            tex.printStackTrace();
            return (IntegerValue)RB.getTciCDRequired().getInteger().newInstance();
        }
	    
	    int nextHeader = ipv6Header.nextHeader.tValue;
        
        if (nextHeader != 135) {
	        RecordOfValue ehrov = (RecordOfValue)p_msg.getField("extHdrList");
	        
	        if (!ehrov.notPresent()) {
                for (int i=0; (i<ehrov.getLength())&&(v_cnt!=p_entry.getInt()); i++) {
	                RecordValue rv = (RecordValue) ((UnionValue)ehrov.getField(i)).getVariant(((UnionValue)ehrov.getField(i)).getPresentVariantName());

	    	        switch (nextHeader) {
	    	        case HOP:
	    	            msgLen -= (((IntegerValue)rv.getField("hdrExtLen")).getInt() * 8 + 8);
	    	            nextHeader = ((IntegerValue)rv.getField("nextHeader")).getInt();
	    	            break;
	    	        case TUNNELEDIPV6:
	    	            msgLen -= ipv6HeaderSize;
	    	            nextHeader = ((IntegerValue)rv.getField("nextHeader")).getInt();
	    	            v_cnt++;
	    	            break;
	    	        case ROUTING:
	    	            msgLen -= (((IntegerValue)rv.getField("hdrExtLen")).getInt() * 8 + 8);
	    	            nextHeader = ((IntegerValue)rv.getField("nextHeader")).getInt();
	    	            break;
	    	        case FRAGMENT:
	    	            msgLen -= 8;
	    	            nextHeader = ((IntegerValue)rv.getField("nextHeader")).getInt();
	    	            break;
	    	        case DESTINATION:
	    	            msgLen -= (((IntegerValue)rv.getField("hdrExtLen")).getInt() * 8 + 8);
	    	            nextHeader = ((IntegerValue)rv.getField("nextHeader")).getInt();
	    	            break;
	    	        }
	            }
	        }
	        
        }

	    IntegerValue ret = (IntegerValue) RB.getTciCDRequired().getTypeForName("LibCommon_BasicTypesAndValues.UInt16").newInstance();

	    if (v_cnt == p_entry.getInt()) {
	    	ret.setInt(msgLen);
	    }
	    else {
            RB.getTciTMProvided().tciError("fx_tunnelledPayloadLength: Error while calculating payload length.\n" +
            		"Reason: Ipv6 packet does not include the amount of tunnelled Ipv6 headers(41)(found " + v_cnt +
            		" occurence(s), requested was " + p_entry.getInt() + " entry)");
	    }
	    
	    return ret;
	}

	/**
	 * @param p_srcAddress
	 * @param p_dstAddress
	 * @param p_payload
	 * @param p_nextHeader
	 * @return
	 */
	public synchronized OctetstringValue fx_calcPayloadChecksum(OctetstringValue p_srcAddress,
			OctetstringValue p_dstAddress, UnionValue p_payload, IntegerValue p_nextHeader) {
		byte [] pseudoHeader = new byte[40];
	    
	    int checksum = 0;
	    
        byte []src = ((OctetString)p_srcAddress).toByteArray();
        System.arraycopy(src, 0, pseudoHeader, 0, 16);
        //16Byte dst
        byte []dst = ((OctetString)p_dstAddress).toByteArray();
        System.arraycopy(dst, 0, pseudoHeader, 16, 16);

        //4Byte upper layer packet length
        byte[] upperLayerLength = new byte[4];
        
        TabularIpv6Payload ipv6Payload = new TabularIpv6Payload(p_payload);
        TabularEncoder enc = new TabularEncoder(new BitArrayOutputStream(), RB);
        
        try {
			ipv6Payload.encode(enc);
		} catch (TabularException e) {
			e.printStackTrace();
			RB.tciTMProvided.tciError("Checksum calculation: " + e.getMessage());
		    return null;
		}
		
        byte[] payload = enc.toByteArray();
		int len = payload.length;
        upperLayerLength[0] = (byte)(len>>24);
        upperLayerLength[1] = (byte)(len>>16);
        upperLayerLength[2] = (byte)(len>>8);
        upperLayerLength[3] = (byte)(len);
        System.arraycopy(upperLayerLength, 0, pseudoHeader, 32, 4);

        //3Byte zero
        byte [] zero_bytes = new byte[] {0x0, 0x0, 0x0};
        System.arraycopy(zero_bytes, 0, pseudoHeader, 36, 3);

        //1Byte next header
        byte [] next_header = new byte[] {(byte)p_nextHeader.getInt() };
        System.arraycopy(next_header, 0, pseudoHeader, 39, 1);

        //checksum pseudo header
        byte []val = new byte[2];
        for (int i=0; i<40; i+=2) {
        	System.arraycopy(pseudoHeader, i, val, 0, 2);
        	checksum += new BigInteger(val).intValue()&0x0000FFFF;
        	checksum = (checksum>>16) + (checksum & 0x0000FFFF);
        }

        //checksum layer 4 header
//        for (int i=0; i<len; i+=2) {
//        	if ((i+2)<=len) {
//        		System.arraycopy(payload, i, val, 0, 2);
//        		checksum += new BigInteger(val).intValue()&0x0000FFFF;
//        		checksum = (checksum>>16) + (checksum & 0x0000FFFF);
//        	}
//        }
//
//        if ((len%2) != 0) {
//        	byte v1 = payload[len - 1];
//        	val[0] = v1;
//        	val[1] = 0x0;
//
//        	checksum += new BigInteger(val).intValue()&0x0000FFFF;
//        	checksum = (checksum>>16) + (checksum & 0x0000FFFF);
//        }
//
//        checksum = (~checksum) & 0x0000FFFF;
//
//	    OctetstringValue ret = (OctetstringValue) RB.getTciCDRequired().getTypeForName("LibCommon_DataStrings.Oct2").newInstance();
//	    ret.setLength(2);
//	    ret.setOctet(0, (int)checksum>>8);
//	    ret.setOctet(1, (int)checksum&0xFF);
//	    
//	    return ret;
        return checksum16BitComplement(
        		(OctetstringValue) RB.getTciCDRequired().getTypeForName("LibCommon_DataStrings.Oct2").newInstance(),
				payload, 
				checksum, 0, len);
	}
	
	/**
	 * This external function returns the encoded IPv6 packet.
	 * 
	 * @param   p_msg 	The IPv6 packet for encoding.
	 * @return  		The encoded packet.
	 */
	public synchronized OctetstringValue fx_encodeMessage(RecordValue p_msg ){
	    
	    TriMessage msg = RB.TestAdapter.getCodec("").encode(p_msg);
	    byte bmsg[] = msg.getEncodedMessage();

		StringBuffer sb = new StringBuffer();
		
		for (int j = 0; j < bmsg.length; j++) {
		    //we choose capital letters
		    sb.append(hex[1].charAt((bmsg[j] >> 4) & 0xf));
		    sb.append(hex[1].charAt((bmsg[j]) & 0xf));
		}		

	    OctetstringValue ret = (OctetstringValue) RB.getTciCDRequired().getOctetstring().newInstance();
	    ret.setLength(bmsg.length);
	    ret.setString("'" + sb + "'O");
	    
	    return ret;
	}

	/**
	 * This external function calculates the checksum for MIPv6 Header.
	 * If HomeAddressOption present, then this external function is called with the
	 * HomeAddress as Source Address
	 * 
	 * @param  p_srcAddr Source Address to be used for Checksum calculation
	 * @param  p_dstAddr MobileHeader
	 * @param  p_mobileHeader MobileHeader
	 * @return MIPv6 Header checksum
	 */
	public synchronized OctetstringValue fx_mipHeaderChecksum(OctetstringValue srcAddress, OctetstringValue dstAddress, RecordValue p_hdr) {
	    byte [] pseudoHeader = new byte[40];
	    
	    int checksum = 0;
	    
	    TabularMobileHeader mipHeader = new TabularMobileHeader(p_hdr);
	    try {
	    	mipHeader.encode(new TabularEncoder(new BitArrayOutputStream(), RB));
        } catch (TabularException tex) {
            RB.getTciTMProvided().tciError(tex.getMessage());
            tex.printStackTrace();
            return (OctetstringValue)RB.getTciCDRequired().getOctetstring().newInstance();
        }
	    
        //16Byte src
//        byte []src = ipv6Header.sourceAddress.tValue;
	    //TODO for debugging reason
        byte []src = ((OctetString)srcAddress).toByteArray();
        System.arraycopy(src, 0, pseudoHeader, 0, 16);
        //16Byte dst
//        byte []dst = ipv6Header.destinationAddress.tValue;
        byte []dst = ((OctetString)dstAddress).toByteArray();
        System.arraycopy(dst, 0, pseudoHeader, 16, 16);
        
        //4Byte upper layer packet length
        byte[] upperLayerLength = new byte[4];
        int len = mipHeader.headerLen.tValue * 8 + 8;
        upperLayerLength[0] = (byte)(len>>24);
        upperLayerLength[1] = (byte)(len>>16);
        upperLayerLength[2] = (byte)(len>>8);
        upperLayerLength[3] = (byte)(len);
        System.arraycopy(upperLayerLength, 0, pseudoHeader, 32, 4);
        
        //3Byte zero
        byte [] zero_bytes = new byte[] {0x0, 0x0, 0x0};
        System.arraycopy(zero_bytes, 0, pseudoHeader, 36, 3);
        
        //1Byte next header
        byte [] next_header = new byte[] {(byte)0x87};//{0x02 };
        System.arraycopy(next_header, 0, pseudoHeader, 39, 1);
        
        //checksum pseudo header
        byte []val = new byte[2];
        for (int i=0; i<40; i+=2) {
        	System.arraycopy(pseudoHeader, i, val, 0, 2);
        	checksum += new BigInteger(val).intValue()&0x0000FFFF;
        	checksum = (checksum>>16) + (checksum & 0x0000FFFF);
        }
        
        //checksum layer 4 header
//        byte [] wholeMessage = RB.getTciEncoding().getCodec("LibIpv6_Interface_TypesAndValues").encode(p_hdr).getEncodedMessage();
//        for (int i=0; i<len; i+=2) {
//        	if ((i+2)<=len) {
//        		System.arraycopy(wholeMessage, i, val, 0, 2);
//        		checksum += new BigInteger(val).intValue()&0x0000FFFF;
//        		checksum = (checksum>>16) + (checksum & 0x0000FFFF);
//        	}
//        }
//        
//        if ((len%2) != 0) {
//        	byte v1 = wholeMessage[len - 1];
//        	val[0] = v1;
//        	val[1] = 0x0;
//        	
//        	checksum += new BigInteger(val).intValue()&0x0000FFFF;
//        	checksum = (checksum>>16) + (checksum & 0x0000FFFF);
//        }
//        
//        checksum = (~checksum) & 0x0000FFFF;
//        
//	    OctetstringValue ret = (OctetstringValue) RB.getTciCDRequired().getTypeForName("LibCommon_DataStrings.Oct2").newInstance();
//	    ret.setLength(2);
//	    ret.setOctet(0, (int)checksum>>8);
//	    ret.setOctet(1, (int)checksum&0xFF);
//	    
//	    return ret;
        return checksum16BitComplement(
        		(OctetstringValue) RB.getTciCDRequired().getTypeForName("LibCommon_DataStrings.Oct2").newInstance(), 
        		RB.getTciEncoding().getCodec("LibIpv6_Interface_TypesAndValues").encode(p_hdr).getEncodedMessage(), 
        		checksum, 0, len);
	}
	
	/**
	 * This external function calculates the length of MIPv6 Header.
	 * 
	 * @param  p_mobileHeader MobileHeader
	 * @return MIPv6 Header length
	 */
	public synchronized IntegerValue fx_mipHeaderLength(RecordValue p_hdr ){
	    TriMessage message = RB.getTciEncoding().getCodec("LibIpv6_Interface_TypesAndValues").encode(p_hdr);
	    
		//  8-bit unsigned integer, representing the length of the Mobility
	    //  Header in units of 8 octets, excluding the first 8 octets.
	    int msgLen = message.getEncodedMessage().length/8-1;

	    IntegerValue ret = (IntegerValue) RB.getTciCDRequired().getTypeForName("LibCommon_BasicTypesAndValues.UInt8").newInstance();

	    ret.setInt(msgLen);
	    
	    return ret;
	}

	/** 
	 * This external function runs the MobileHeader through Codec and returns 
	 * the octetstring representing the MobileHeader
	 *			
	 * @param 	p_mobileHeader Mip Header to be encoded to octetstring
	 * @return 	octetstring
	 */
	public synchronized OctetstringValue fx_mipHdrToOct(RecordValue p_mipHeader) {

		OctetString o = new OctetString(RB);

		TabularEncoder enc = new TabularEncoder(new BitArrayOutputStream(), RB);

		TabularMobileHeader mh = new TabularMobileHeader(p_mipHeader);
		try {
			mh.encode(enc);
		} catch (TabularException e) {
			System.err.println(e.getMessage());
			e.printStackTrace();
			return o;
		}

		byte[] msg = enc.toByteArray();
		o.setOctetstring(msg);
		o.setLength(msg.length);
		
		return o;
	}
	
	/** 
	 * Apply indicated Integrity algorithm to the message. Message is an octetstring.
	 *			If e_sha1_96 is chosen, then the key input shall be ignored.
	 *			
	 * @param  p_integrityAlgo Cryptographic function used to compute MAC
	 * @param  p_key Key used to compute the MAC
	 * @param  p_message Octetstring message
	 * @return Message authenticator
	 */
	public synchronized OctetstringValue fx_mac(EnumeratedValue p_integrityAlgo, OctetstringValue p_key, OctetstringValue p_message) {

		String e = p_integrityAlgo.getEnum();
//		System.out.println("fx_mac: input: " + p_message.getString());
		return integrity(e, p_key, p_message);
	}

	/**
	 * This external function sets the test adapter with the relevant security parameters
	 *			
	 * @param 	p_sa Security Association
	 * @return 	FncRetCode
	 */
	public synchronized EnumeratedValue fx_setSecurityParameters(RecordOfValue p_sa) {
		EnumeratedValue fncRetCode = (EnumeratedValue)RB.getTciCDRequired().getTypeForName("LibCommon_VerdictControl.FncRetCode").newInstance();
		fncRetCode.setEnum("e_success");
		
		for (int i = 0; i < p_sa.getLength(); i++) {
			if (!p_sa.getField(i).notPresent()) {
				((IPv6TestAdapter)RB.TestAdapter).setSecurityInfo(
						((IntegerValue)((RecordValue)p_sa.getField(i)).getField("spi")).getBigInt(),
						((IntegerValue)((RecordValue)p_sa.getField(i)).getField("seqNr")).getInt(),
						((EnumeratedValue)((RecordValue)p_sa.getField(i)).getField("ahIntegrityAlgo")),
						((OctetString)((RecordValue)p_sa.getField(i)).getField("ahIntegrityKey")),
						((IntegerValue)((RecordValue)p_sa.getField(i)).getField("ahIcvLen")).getInt(),
						((EnumeratedValue)((RecordValue)p_sa.getField(i)).getField("espEncryptionAlgo")),
						((OctetString)((RecordValue)p_sa.getField(i)).getField("espEncryptionKey")),
						((OctetString)((RecordValue)p_sa.getField(i)).getField("espIv")),
						((IntegerValue)((RecordValue)p_sa.getField(i)).getField("espEncrBlockSize")).getInt(),
						((EnumeratedValue)((RecordValue)p_sa.getField(i)).getField("espIntegrityAlgo")),
						((OctetString)((RecordValue)p_sa.getField(i)).getField("espIntegrityKey")),
						((IntegerValue)((RecordValue)p_sa.getField(i)).getField("espIcvLen")).getInt(),
//						((EnumeratedValue)((RecordValue)p_sa.getField(i)).getField("espCombinedModeAlgo")),
//						((OctetString)((RecordValue)p_sa.getField(i)).getField("espCombinedModeKey")),
						((EnumeratedValue)((RecordValue)p_sa.getField(i)).getField("ipSecProtocolMode"))
				);
			}
		}
		
		return fncRetCode;
	}

	/**
	 * This external function sets the test adapter with the relevant security parameters
	 *			
	 * @param 	p_sa Security Association
	 * @return 	FncRetCode
	 */
	public synchronized EnumeratedValue fx_setIkeSecurityParameters(RecordValue p_ikeSa) {
		EnumeratedValue fncRetCode = (EnumeratedValue)RB.getTciCDRequired().getTypeForName("LibCommon_VerdictControl.FncRetCode").newInstance();
		fncRetCode.setEnum("e_success");
		
		((IPv6TestAdapter)RB.TestAdapter).setIKESecurityInfo(
				(OctetString)p_ikeSa.getField("spiInitiator"),
				(OctetString)p_ikeSa.getField("spiResponder"),
//				(IntegerValue)p_ikeSa.getField("messageID"),
				(EnumeratedValue)p_ikeSa.getField("ikeEncryptionAlgo"),
//				(OctetString)p_ikeSa.getField("ikeEncryptionKey"),
				((IntegerValue)p_ikeSa.getField(IKE_ENCR_KEY_LEN)).getInt(),
				((IntegerValue)p_ikeSa.getField("ikeEncrBlockSize")).getInt(),
				(EnumeratedValue)p_ikeSa.getField("ikeIntegrityAlgo"),
//				(OctetString)p_ikeSa.getField("ikeIntegrityKey"),
				((IntegerValue)p_ikeSa.getField(IKE_INTEGR_KEY_LEN)).getInt(),
				((IntegerValue)p_ikeSa.getField("ikeIcvLen")).getInt(),
				(EnumeratedValue)p_ikeSa.getField("ikePseudoRandomFunction"),
				((IntegerValue)p_ikeSa.getField(IKE_PRFKEY_LEN)).getInt(),
				(EnumeratedValue)p_ikeSa.getField("diffieHellmanGroup"),
				(OctetString)p_ikeSa.getField("diffieHellmanPrivKey"),
				(OctetString)p_ikeSa.getField("diffieHellmanSharedSecret"),
				(RecordValue)p_ikeSa.getField("sevenSecrets")
		);
		
		return fncRetCode;
	}

	/**
	 * This external function calculates the payload length
	 * of an IPv6 authentication header
	 * @param   p_authHeader Authentication Header
	 * @return  payload length in 32-bits words
	*/
	public synchronized IntegerValue fx_authHeaderPayloadLength(RecordValue p_authHeader) {
		IntegerValue payloadLength = (IntegerValue)RB.tciCDRequired.getInteger().newInstance();
	    int length = ((OctetstringValue)p_authHeader.getField("icv")).getLength() + 12;
	    int tmp = 0;
	    //align to 8-octet units
	    tmp = length/8;
	    if (length%8 != 0) {
	    	++tmp;
	    }
		payloadLength.setInt(tmp*2 - 2);
	    
		return payloadLength;
	}

	/**
	 * This external function implements the negotiated
	 * pseudo random function and calcutes pseudo random value
	 * based on the two input strings.
	 * @param   p_prf negotiated PseudoRandomFunction
	 *			choice is:
	 *			e_prfHmacMd5(1),
	 *			e_prfHmacSha1(2),
	 *			e_prfHmacTiger(3),
	 *			e_prfAes128Xcbc(4)
	 * @param   p_first, p_second input values to pseudo random function
	 * @return  pseudo random value 
	*/
	public synchronized OctetstringValue fx_pseudoRandom(EnumeratedValue p_prf,
									   OctetstringValue p_first,
									   OctetstringValue p_second ) {

		byte [] prv = innerPRF(p_prf.getEnum(), ((OctetString)p_first).toByteArray(), ((OctetString)p_second).toByteArray());
		
		return new OctetString(RB, prv.length, prv);
	}

	/**
	 * This external function implements the negotiated
	 * pseudo random function and calcutes the seven secrets
	 * based on the two input strings.
	 * SK_d  = prf (p_first, S | 0x01)
   	 * SK_ai = prf (p_first, SK_d  | p_second | 0x02)
   	 * SK_ar = prf (p_first, SK_ai | p_second | 0x03)
  	 * SK_ei = prf (p_first, SK_ar | p_second | 0x04)
  	 * SK_er = prf (p_first, SK_ei | p_second | 0x05)
  	 * SK_pi = prf (p_first, SK_er | p_second | 0x06)
  	 * SK_pr = prf (p_first, SK_pi | p_second | 0x07)
	 * @param   p_prf negotiated PseudoRandomFunction
	 *			choice is:
	 *			e_prfHmacMd5(1),
	 *			e_prfHmacSha1(2),
	 *			e_prfHmacTiger(3),
	 *			e_prfAes128Xcbc(4)
	 * @param   p_first, p_second input values to pseudo random function
	 * @return  pseudo random value 
	*/
//	public synchronized RecordValue fx_pseudoRandomPlus(EnumeratedValue p_prf,
	public synchronized RecordValue fx_pseudoRandomPlus(RecordValue p_ikeSa,
									  	   OctetstringValue p_first,
									   	   OctetstringValue p_second ) {

		//reset the static variables used for in getKey
		block = null;
		blockPointer = 0;

		RecordValue sevenSecrets = (RecordValue)RB.tciCDRequired.getTypeForName("LibIpv6_Rfc4306Ikev2_TypesAndValues.SevenSecrets").newInstance();

		byte [] first = ((OctetString)p_first).toByteArray();
		byte [] second = ((OctetString)p_second).toByteArray();
		
		String prf = ((EnumeratedValue)p_ikeSa.getField("ikePseudoRandomFunction")).getEnum();

		IntegerValue encrKeyLengthIv = (IntegerValue)p_ikeSa.getField(IKE_ENCR_KEY_LEN);
		if (encrKeyLengthIv.notPresent()) {
			RB.tciTMProvided.tciError("fx_pseudoRandomPlus: ikeEncryptionKeyLength not set!");
			return null;
		}
		int encrKeyLength = encrKeyLengthIv.getInt();
		IntegerValue integrityKeyLengthIv = (IntegerValue)p_ikeSa.getField(IKE_INTEGR_KEY_LEN);
		if (integrityKeyLengthIv.notPresent()) {
			RB.tciTMProvided.tciError("fx_pseudoRandomPlus: ikeIntegrityKeyLength not set!");
			return null;
		}
		int integrityKeyLength = integrityKeyLengthIv.getInt();
		IntegerValue prfKeyLengthIv = (IntegerValue)p_ikeSa.getField(IKE_PRFKEY_LEN);
		if (prfKeyLengthIv.notPresent()) {
			RB.tciTMProvided.tciError("fx_pseudoRandomPlus: ikePRFKeyLength not set!");
			return null;
		}
		int prfKeyLength = prfKeyLengthIv.getInt();
		
		try {
			ByteArrayOutputStream key = new ByteArrayOutputStream();
			int cnt = 0;
			//sK_d
			cnt = getKey(cnt, prf, first, second, key, 16);
			OctetString sK_d = new OctetString(RB, key.size(), key.toByteArray());
			key.reset();
			sevenSecrets.setField("sK_d", sK_d);
			//sK_ai
//			cnt = getKey(cnt, prf, first, second, key, getKeySize(((IPv6TestAdapter)RB.TestAdapter).getIkeIntegrityAlgo()));
//			cnt = getKey(cnt, prf, first, second, key, ((IPv6TestAdapter)RB.TestAdapter).getIkeIntegrityKeyLength());
			cnt = getKey(cnt, prf, first, second, key, integrityKeyLength);
			OctetString sK_ai = new OctetString(RB, key.size(), key.toByteArray());
			key.reset();
			sevenSecrets.setField("sK_ai", sK_ai);
			//sK_ar
//			cnt = getKey(cnt, prf, first, second, key, getKeySize(((IPv6TestAdapter)RB.TestAdapter).getIkeIntegrityAlgo()));
//			cnt = getKey(cnt, prf, first, second, key, ((IPv6TestAdapter)RB.TestAdapter).getIkeIntegrityKeyLength());
			cnt = getKey(cnt, prf, first, second, key, integrityKeyLength);
			OctetString sK_ar = new OctetString(RB, key.size(), key.toByteArray());
			key.reset();
			sevenSecrets.setField("sK_ar", sK_ar);
			//sK_ei
//			cnt = getKey(cnt, prf, first, second, key, getKeySize(((IPv6TestAdapter)RB.TestAdapter).getIkeEncryptionAlgo()));
//			cnt = getKey(cnt, prf, first, second, key, ((IPv6TestAdapter)RB.TestAdapter).getIkeEncryptionKeyLength());
			cnt = getKey(cnt, prf, first, second, key, encrKeyLength);
			OctetString sK_ei = new OctetString(RB, key.size(), key.toByteArray());
			key.reset();
			sevenSecrets.setField("sK_ei", sK_ei);
			//sK_er
//			cnt = getKey(cnt, prf, first, second, key, getKeySize(((IPv6TestAdapter)RB.TestAdapter).getIkeEncryptionAlgo()));
//			cnt = getKey(cnt, prf, first, second, key, ((IPv6TestAdapter)RB.TestAdapter).getIkeEncryptionKeyLength());
			cnt = getKey(cnt, prf, first, second, key, encrKeyLength);
			OctetString sK_er = new OctetString(RB, key.size(), key.toByteArray());
			key.reset();
			sevenSecrets.setField("sK_er", sK_er);
			//sK_pi
			cnt = getKey(cnt, prf, first, second, key, prfKeyLength);
			OctetString sK_pi = new OctetString(RB, key.size(), key.toByteArray());
			key.reset();
			sevenSecrets.setField("sK_pi", sK_pi);
			//sK_pr
			cnt = getKey(cnt, prf, first, second, key, prfKeyLength);
			OctetString sK_pr = new OctetString(RB, key.size(), key.toByteArray());
			key.reset();
			sevenSecrets.setField("sK_pr", sK_pr);
			
			//debug statements
			System.out.println("sK_d" + HexViewer.hexString(sK_d.toByteArray() ) + "\n");
			System.out.println("sK_ai" + HexViewer.hexString(sK_ai.toByteArray()) + "\n");
			System.out.println("sK_ar" + HexViewer.hexString(sK_ar.toByteArray()) + "\n");
			System.out.println("sK_ei" + HexViewer.hexString(sK_ei.toByteArray()) + "\n");
			System.out.println("sK_er" + HexViewer.hexString(sK_er.toByteArray()) + "\n");
			System.out.println("sK_pi" + HexViewer.hexString(sK_pi.toByteArray()) + "\n");
			System.out.println("sK_pr" + HexViewer.hexString(sK_pr.toByteArray()) + "\n");
			
		} catch (IOException e) {
			RB.tciTMProvided.tciError(e.getMessage());
			e.printStackTrace();
		}
		
		return sevenSecrets;
	}
	
	 /**
	  * @desc This external function implements the negotiated
	  *          pseudo random function and calcutes pseudo random value
	  *   based on the two input strings.
	  * @param   p_prf negotiated PseudoRandomFunction
	  *   choice is:
	  *   e_prfHmacMd5(1),
	  *   e_prfHmacSha1(2),
	  *   e_prfHmacTiger(3),
	  *   e_prfAes128Xcbc(4)
	  * @param   p_first, p_second input values to pseudo random function
	  * @param   p_EncIntKeys input/output value carries a record of
	  *          encryption and integrity key lengths and key values
	  *          Input value contains only the key lengths,
	  *          length values set to 0 indicate that no correspondent keys need to be calculated.
	  *          Output value contains key values taken from KEYMAT
	  *          according to the rules of RFC4306, clause 2.17.
	 */
	public synchronized EnumeratedValue fx_calculateKEYMATandDistribute (EnumeratedValue p_prf,
			OctetstringValue p_first,
			OctetstringValue p_second,
			RecordValue p_encIntKeys) {
		EnumeratedValue fncRetCode = (EnumeratedValue)RB.getTciCDRequired().getTypeForName("LibCommon_VerdictControl.FncRetCode").newInstance();
		fncRetCode.setEnum("e_success");

//		reset the static variables used for in getKey
		block = null;
		blockPointer = 0;

		byte [] first = ((OctetString)p_first).toByteArray();
		byte [] second = ((OctetString)p_second).toByteArray();

		String prf = p_prf.getEnum();

		IntegerValue encrKeyLength = (IntegerValue)p_encIntKeys.getField("encKeyLength");
		IntegerValue intKeyLength = (IntegerValue)p_encIntKeys.getField("intKeyLength");

		try {
			ByteArrayOutputStream key = new ByteArrayOutputStream();
			int cnt = 0;
//			encKeyResponder
			if (encrKeyLength.getInt() > 0) {
				cnt = getKey(cnt, prf, first, second, key, encrKeyLength.getInt());
				OctetString encKeyResponder = new OctetString(RB, key.size(), key.toByteArray());
				key.reset();
				p_encIntKeys.setField("encKeyResponder", encKeyResponder);
			}
//			intKeyResponder
			if (intKeyLength.getInt() > 0) {
				cnt = getKey(cnt, prf, first, second, key, intKeyLength.getInt());
				OctetString intKeyResponder = new OctetString(RB, key.size(), key.toByteArray());
				key.reset();
				p_encIntKeys.setField("intKeyResponder", intKeyResponder);
			}
//			encKeyInitiator
			if (encrKeyLength.getInt() > 0) {
				cnt = getKey(cnt, prf, first, second, key, encrKeyLength.getInt());
				OctetString encKeyInitiator = new OctetString(RB, key.size(), key.toByteArray());
				key.reset();
				p_encIntKeys.setField("encKeyInitiator", encKeyInitiator);
			}
//			intKeyInitiator
			if (intKeyLength.getInt() > 0) {
				cnt = getKey(cnt, prf, first, second, key, intKeyLength.getInt());
				OctetString intKeyInitiator = new OctetString(RB, key.size(), key.toByteArray());
				key.reset();
				p_encIntKeys.setField("intKeyInitiator", intKeyInitiator);
			}
		} catch (IOException e) {
			RB.tciTMProvided.tciError(e.getMessage());
			e.printStackTrace();
		}

		return fncRetCode;
	}

	/**
	 * This external function implements the Diffie-Hellman procedure
	 * and calculates the key for the Key Exchange payload to be sent to the 
	 * peer based on the private key and the number of the Diffie-Hellman group.
	 * Prime and generator for the Diffie-Hellman groups are found in:
	 * Group 1 and 2: RFC4306 Appendix B
	 * Group 5, 14, 15, 16, 17 and 18: RFC3526
	 * @param   p_diffieHellmanGroup chosen Diffie-Hellman group
	 * @param   p_diffieHellmanPrivKey private key taken from PIXIT value
	*/
	public synchronized OctetstringValue fx_dHKeyToSend(EnumeratedValue p_diffieHellmanGroup,
									  OctetstringValue p_diffieHellmanPrivKey ) {
		return calcDH(p_diffieHellmanGroup, p_diffieHellmanPrivKey, new BigInteger("2"));
	}

	/**
	 * This external function implements the Diffie-Hellman procedure
	 * and calculates the shared secret based on the private key,
	 * the received key from the Key Exchange payload and the number
	 * of the Diffie-Hellman group.
	 * Prime and generator for the Diffie-Hellman groups are found in:
	 * Group 1 and 2: RFC4306 Appendix B
	 * Group 5, 14, 15, 16, 17 and 18: RFC3526
	 * @param   p_diffieHellmanGroup chosen Diffie-Hellman group
	 * @param   p_diffieHellmanPrivKey private key taken from PIXIT value
	 * @return  p_receivedKey received key taken from Key Exchange payload 
	*/
	public synchronized OctetstringValue fx_dHSharedSecret(EnumeratedValue p_diffieHellmanGroup,
									     OctetstringValue p_diffieHellmanPrivKey,
									     OctetstringValue p_receivedKey) {
		return calcDH(p_diffieHellmanGroup, p_diffieHellmanPrivKey, new BigInteger(1, ((OctetString) p_receivedKey).toByteArray()));		
	}

	/**
	 * This external function runs the Ipv6Payload through Codec and returns 
	 * the octetstring representing the Ipv6Payload
	 *			
	 * @param 	p_ipv6Payload Ipv6Payload to be encoded to octetstring
	 * @return 	octetstring
	 */
	public synchronized OctetstringValue fx_ipv6PayloadToOct(UnionValue p_ipv6Payload) {

		OctetString o = new OctetString(RB);

		TabularEncoder enc = new TabularEncoder(new BitArrayOutputStream(), RB);

		TabularIpv6Payload ip = new TabularIpv6Payload(p_ipv6Payload);
		try {
			ip.encode(enc);
		} catch (TabularException e) {
			System.err.println(e.getMessage());
			e.printStackTrace();
			return o;
		}

		byte[] msg = enc.toByteArray();
		o.setOctetstring(msg);
		o.setLength(msg.length);
		
		return o;
	}

	/**
	 * This external function calculates the payload length
	 * of the IKE encrypted payload. Before calculating the lenght, the payload is encrypted in the test adapter.
	 * @param   p_msg Encrypted Payload
	 * @return  payload length in bytes
	 */
	public synchronized IntegerValue fx_ikeEncPayloadLength(RecordValue p_msg) {

		IntegerValue rc = (IntegerValue)RB.tciCDRequired.getTypeForName("LibCommon_BasicTypesAndValues.UInt16").newInstance();

		RecordValue encPayload = null;

		UnionValue ipv6Payload = (UnionValue) p_msg.getField("ipv6Payload");
		if (ipv6Payload.getPresentVariantName().equals("ikeMsg")) {
			RecordValue ikeMsg = (RecordValue) ipv6Payload.getVariant("ikeMsg");
			
			RecordOfValue payloadList = (RecordOfValue) ikeMsg.getField("payloadList");
			for (int i=0; i<payloadList.getLength();i++) {
				UnionValue payload = ((UnionValue)payloadList.getField(i));
				if (payload.getPresentVariantName().equals("encrypted")) {
					encPayload = (RecordValue) payload.getVariant("encrypted");
					break;
				}
			}
		}
		
		if (encPayload == null) {
			RB.tciTMProvided.tciError("fx_ikeEncPayloadLength: Function was called with an invalid message! No IKE message including encryptedPayload given!");
		}
		
		TabularEncoder enc = new TabularEncoder(new BitArrayOutputStream(), RB);

		TabularEncryptedPayload ep = new TabularEncryptedPayload(encPayload);
		try {
			ep.encode(enc);
		} catch (TabularException e) {
			System.err.println(e.getMessage());
			e.printStackTrace();
			return rc;
		}

		byte[] msg = enc.toByteArray();
		
		rc.setInt(msg.length);
		
		return rc;
	}

	/**
	 * This external function runs the IkePayloadList through Codec and returns 
	 * the octetstring representing the IkePayloadList. If encrypted payload is part of the payload list,
	 * then the ecncrypted octet string is returned.
	 *			
	 * @param 	p_ikev2Header Ikev2Header to be encoded to octetstring
	 * @return 	octetstring
	 */
	public synchronized OctetstringValue fx_ikePayloadListToOct(RecordOfValue p_ikePayloadList) {

		OctetString o = new OctetString(RB);

		TabularEncoder enc = new TabularEncoder(new BitArrayOutputStream(), RB);

		TabularIkePayloadList ipl = new TabularIkePayloadList(p_ikePayloadList);
		try {
			ipl.encode(enc);
		} catch (TabularException e) {
			System.err.println(e.getMessage());
			e.printStackTrace();
			return o;
		}

		byte[] msg = enc.toByteArray();
		o.setOctetstring(msg);
		o.setLength(msg.length);
		
		return o;
	}

	/**
	 * This external function runs the Ikev2Header through Codec and returns 
	 * the octetstring representing the Ikev2Header
	 *			
	 * @param 	p_ikev2Header Ikev2Header to be encoded to octetstring
	 * @return 	octetstring
	 */
	public synchronized OctetstringValue fx_ikev2HeaderToOct(RecordValue p_ikev2Header) {

		OctetString o = new OctetString(RB);

		TabularEncoder enc = new TabularEncoder(new BitArrayOutputStream(), RB);

		TabularIkev2Header ih = new TabularIkev2Header(p_ikev2Header);
		try {
			ih.encode(enc);
		} catch (TabularException e) {
			System.err.println(e.getMessage());
			e.printStackTrace();
			return o;
		}

		byte[] msg = enc.toByteArray();
		o.setOctetstring(msg);
		o.setLength(msg.length);
		
		return o;
	}
	
	/////////////////////////////////////////////////
	// TODO                                        //
	// unused functions, please check for deletion //
	// be green                                    //
	/////////////////////////////////////////////////
	/**
	 * This external function calculates the checksum
	 * 
	 * @param  p_packet The Icmpv6 packet for calculation. 
	 * @return 			The calculated checksum.
	 */
//	TODO check if it is used	
	public synchronized OctetstringValue fx_icmpv6Checksum(RecordValue p_packet) {
		
	    return calcChecksum(p_packet, 58);
	}

	/** This external function calculates a Message Authentication Code
	 * @param  p_key Key used to compute the MAC
	 * @param  p_message Octetstring message
	 * @param  p_crypto Cryptographic function used to compute MAC
	 * @return Message HMAC
	*/
//TODO check if it is used	
	public synchronized OctetstringValue fx_computeHMAC(OctetstringValue p_key, OctetstringValue p_message, EnumeratedValue p_crypto ) {
	    OctetstringValue ret = (OctetstringValue) RB.getTciCDRequired().getOctetstring().newInstance();
	    return ret;
	}

	/** This external function calculates a SHA1 hash
	 * @param  p_message Octetstring message
	 * @param  p_crypto Cryptographic function used to compute MAC
	 * @return Message hash
	*/
//	TODO check if it is used	
	public synchronized OctetstringValue fx_computeHash(OctetstringValue p_key, OctetstringValue p_message, EnumeratedValue p_crypto ) {
	    OctetstringValue ret = (OctetstringValue) RB.getTciCDRequired().getOctetstring().newInstance();
	    return ret;
	}

	/** This external function computes the Authenticator for 
	 *          return routability procedure
	 * @param 	p_careOfAddr Care-of address
	 * @param 	p_cnAddr Address of the correspondant
	 * @param 	p_ipv6Packet Ipv6 Packet
	 * @param 	p_kbm Binding management key
	 * @return 	Authenticator
	*/
//	TODO check if it is used	
	public synchronized OctetstringValue fx_bindingAuthenticator(OctetstringValue p_careOfAddrHae, 
			OctetstringValue p_careOfAddr, 
			OctetstringValue cnAddr,
			OctetstringValue p_ipv6Packet,
			OctetstringValue p_kbm) {
	    OctetstringValue ret = (OctetstringValue) RB.getTciCDRequired().getOctetstring().newInstance();
	    return ret;
	}
	
	/** This external function computes Integrity Check Value
	 * @param 	p_ipv6Packet Ipv6 Packet
	 * @param 	p_sa Security association
	 * @return 	Integrity Check Value
	*/
//	TODO check if it is used	
	public synchronized OctetstringValue fx_IntegrityCheckValue(UnionValue p_ipv6Packet,
			OctetstringValue p_sa) {
	    OctetstringValue ret = (OctetstringValue) RB.getTciCDRequired().getOctetstring().newInstance();
	    return ret;
	}

	/** @desc 	This external function encrypts the EspPayload(except iv) + padding + padLen + nextHeader 
	 *
	 * @param 	p_spi Security parameter Index whihc poinst to the SAD to be used
	 * @param 	p_espHeader ESP Header to be encoded to octetstring
	 * @return 	octetstring The ciphered data
	 */
//	TODO check if it is used	
	public synchronized OctetstringValue fx_encrypt(IntegerValue p_spi, RecordValue p_espHeader) {

		IPv6TestAdapter ta = ((IPv6TestAdapter)RB.TestAdapter);

		TabularEncoder encPayload = new TabularEncoder(new BitArrayOutputStream(), RB); 
		TabularEspPayload espPayload = (TabularEspPayload) new TabularEspPayload(p_espHeader.getField(
		"espPayload"));
		
		try {
			Value iv = ((RecordValue)espPayload.tciValue).getField("iv");
			espPayload.iv = (TabularOctetstring) new TabularOctetstring(iv);
			if (!iv.notPresent()) {			
				espPayload.iv.tValue = ((OctetString)espPayload.iv.tciValue).toByteArray();
			}
			
			// -- espIpDatagram
			espPayload.espIpDatagram = (TabularEspIpDatagram) new TabularEspIpDatagram(((RecordValue)espPayload.tciValue).getField("espIpDatagram"));
			espPayload.espIpDatagram.encode(encPayload);
			
			OctetstringValue paddingValue = (OctetstringValue)p_espHeader.getField("padding");
			IntegerValue padLengthValue = (IntegerValue)p_espHeader.getField("padLength");

			boolean genPadding = true;
			OctetstringValue pad = fx_generatePadding(p_spi, (RecordValue)espPayload.tciValue);
			if (pad.getLength() == padLengthValue.getInt()) {
				if (padLengthValue.getInt() == paddingValue.getLength()) {
					genPadding = false;
					// -- padding
					TabularOct1to255 padding = (TabularOct1to255) new TabularOct1to255(paddingValue);
					padding.encode(encPayload);
					
					// -- padLength
					TabularUInt8 padLength = (TabularUInt8) new TabularUInt8(padLengthValue);
					padLength.encode(encPayload);
				}
			}
			if (genPadding) {
				String warning = "encodeESPHeader: Incorrect Padding! Replacing padding!";
				System.out.println(warning);
				RB.getTciTLProvided().tliRT(null, System.nanoTime(), null, 0, null, RTLoggingConstants.RT_LOG_WARN, warning);
				OctetString o = (OctetString)fx_generatePadding(p_spi, (RecordValue)espPayload.tciValue);
				// -- padding
				byte[] oByteArray = o.toByteArray();
				encPayload.encodeOctetstring(oByteArray);
				// -- padLength
				encPayload.encodeByte((byte)oByteArray.length);
			}
			
			// -- nextHdr
			TabularUInt8 nextHdr = (TabularUInt8) new TabularUInt8(p_espHeader.getField("nextHdr"));
			nextHdr.encode(encPayload);
		} catch (TabularException e) {
			RB.tciTMProvided.tciError(e.getMessage());
		}

		byte [] payload = encPayload.toByteArray();
		if (!ta.getEspEncryptionAlgo(p_spi.getBigInt()).equals(ENCRYPTION_E_NULL)) {
			ByteArrayOutputStream ivBaos = new ByteArrayOutputStream();
			if (espPayload.iv.tValue != null) {
				try {
					ivBaos.write(espPayload.iv.tValue);
				} catch (IOException e) {
					RB.tciTMProvided.tciError(e.getMessage());
				}
			}
			payload = Encryption.getEncryptedString(RB, ta.getEspEncryptionAlgo(p_spi.getBigInt()).getEnum(), ta.getEspEncryptionKeyAlgo(p_spi.getBigInt()), ((OctetString)ta.getEspEncryptionKey(p_spi.getBigInt())).toByteArray(), payload, ivBaos);
		}
		
		return new OctetString(RB, payload.length, payload);		
	}

	/** 
	 * Apply indicated Integrity algorithm to the message. Message can only be an EspHeader.
	 *			If e_sha1_96 is chosen, then the key input shall be ignored.
	 *			
	 * @param  p_integrityAlgo Cryptographic function used to compute MAC
	 * @param  p_key Key used to compute the MAC
	 * @param  p_espHeader EspHeader
	 * @param  p_integrityResult  Result of Integrity function
	 * @return Message HMAC
	 */
/* will be deleted in TTCN-3, added fx_espHdrToOct to be able to use one integrity function fx_integrity
external function fx_integrityEspHdr( 	in IntegrityAlgo p_integrityAlgo,
					in octetstring p_key,
					in EspHeader p_espHeader,
					out IntegrityResult p_integrityResult) 
return FncRetCode;
*/
	
	/** 
	 * This external function runs the EspHeader through Codec and returns 
	 *			the octetstring representing the EspHeader
	 *			
	 * @param 	p_mipHeader MIp Header to be encoded to octetstring
	 * @return 	octetstring
	 */
//TODO check if it's used	
	public synchronized OctetstringValue fx_espHdrToOct(RecordValue p_espHeader) {

		OctetString o = new OctetString(RB);

		TabularEncoder enc = new TabularEncoder(new BitArrayOutputStream(), RB);

		TabularEspHeader eh = new TabularEspHeader(p_espHeader);
		try {
			eh.encode(enc);
		} catch (TabularException e) {
			System.err.println(e.getMessage());
			e.printStackTrace();
			return o;
		}

		byte[] msg = enc.toByteArray();
		o.setOctetstring(msg);
		o.setLength(msg.length);
		
		return o;
	}
	
	/** 
	 * This external function computes Integrity Check Value
	 * @param 	p_ipv6Packet Ipv6 Packet
	 * @param 	p_sa Security association
	 * @return 	Integrity Check Value
	 */
/*	
external function fx_integrityCheckValue( in Ipv6Packet p_ipv6Packet,
					  in octetstring p_sa,
					  out octetstring p_result)
return FncRetCode;
*/
	
	/** 
	 * This external function byte aligns the EspPayload (Iv, EspIpDatagram, TfcPadding) + next Header + pad Length
	 *			with padding according to the byteBoundary set by the encryption algo.
	 *			The nextHeader is of constant size 1 octet,
	 *			and is therefore no input parameter to this external function.
	 *			The padLnength is of constant size 1 octet,
	 *			and is therefore no input parameter to this external function.
	 *			
	 * @param 	p_spi Security parameter Index whihc poinst to the SAD to be used
	 * @param 	p_espPayload EspPayload to be byteAligned
	 * @return 	octetstring
	 */
//	TODO check if it is used	
	public synchronized OctetstringValue fx_generatePadding(IntegerValue p_spi, RecordValue p_espPayload) {
		TabularEncoder enc = new TabularEncoder(new BitArrayOutputStream(), RB);
		TabularEspPayload espPayload = new TabularEspPayload(p_espPayload);
		if (!espPayload.tciValue.notPresent()) {
			try {
				espPayload.encode(enc);
			} catch (TabularException e) {
				RB.tciTMProvided.tciError(e.getMessage());
				e.printStackTrace();
			}
		}
		
		BigInteger spi = p_spi.getBigInt();
		String algo = ((IPv6TestAdapter)RB.TestAdapter).getEspEncryptionAlgo(spi).getEnum();//EncryptionAlgo_E_TRIPLEDES_CBC;
		
		int blockSize = 4;
		if (algo != null) {
			if (algo.equals(ENCRYPTION_E_NULL)) {
				//finally it's 1, however I just say 4
				blockSize = 4;
			}
			else if (algo.equals(ENCRYPTION_E_TRIPLEDES_CBC)) {
				blockSize = 8;
			}
			else if (algo.equals(ENCRYPTION_E_AES_CBC)) {
				blockSize = 16;
			}
			else if (algo.equals(ENCRYPTION_E_AES_CTR)) {
				blockSize = 16;
			}
			else if (algo.equals(ENCRYPTION_E_DES_CBC)) {
				blockSize = 8;
			}
		}
		else {
			RB.tciTMProvided.tciError("fx_generatePadding: Encryption algorithm not yet configured!");
			return null;
		}
		int len = blockSize-(enc.toByteArray().length+2)%blockSize;//2 - 1padLength+1nextHdr
		int cnt = 1;
		byte [] ba = new byte[len];
		for (int i=0; i<len; i++) {
			ba[i] = (byte)cnt++;
		}

		return new OctetString(RB, ba.length, ba);
		
//		EnumeratedValue fncRetCode = (EnumeratedValue)RB.getTciCDRequired().getTypeForName("LibCommon_VerdictControl.FncRetCode").newInstance();
//		fncRetCode.setEnum("e_success");
//		
//		int bound = p_byteBoundary.getInt();
//		if (bound != 0) {
//			OctetstringValue ov = fx_espHdrToOct(p_espHeader);
//			int mod = ov.getLength()%bound;
//			if (mod != 0) {
//				((OctetString)p_result).setOctetstring(new byte[bound-mod]);
//			}
//			else {
//				((OctetString)p_result).setOmit();
//			}
//		}
//		else {
//			RB.getTciTLProvided().tliRT(null, System.nanoTime(), null, 0, null, RTLoggingConstants.RT_LOG_ERROR, "byteBoundary cannot be 0!");
//			fncRetCode.setEnum("e_error");
//		}
//		
//		return fncRetCode;
	}

	/** 
	 * Apply indicated Integrity algorithm to the message. Message is an octetstring.
	 *			If e_sha1_96 is chosen, then the key input shall be ignored.
	 *			
	 * @param  p_spi Security parameter Index which poinst to the SAD to be used
	 * @param  p_key Key used to compute the MAC
	 * @param  p_message Octetstring message
	 * @return ICV
	 */
//	TODO check if it is used	
	public synchronized OctetstringValue fx_integrity(IntegerValue p_spi, OctetstringValue p_key, OctetstringValue p_message) {

		String e = ((IPv6TestAdapter)RB.TestAdapter).getEspIntegrityAlgo(p_spi.getBigInt()).getEnum();
		
//		System.out.println("fx_integrity: input: " + p_message.getString());
		OctetstringValue result = null;

		if (e.equals(INTEGRITY_E_AUTH_NULL)) {
			result = p_message;
		}
		else if (e.equals(INTEGRITY_E_AUTH_HMACSHA1_96)) {
			result = Encryption.getEncryptHMACx(RB, "HmacSHA1", p_key, p_message, 96);
		}
		else if (e.equals(INTEGRITY_E_AUTH_AESXCBC_MAC_96)) {
			RB.tciTMProvided.tciError("fx_integrity: algorithm " + e + " not yet implemented!");
			result = null;
		}
		else if (e.equals(INTEGRITY_E_AUTH_HMACMD5_96)) {
			result = Encryption.getEncryptHMACx(RB, "HmacMD5", p_key, p_message, 96);
		}
		else if (e.equals(INTEGRITY_E_AUTH_SHA1)) {
			result = Encryption.getSHA1Integrity(RB, p_message, 160);
		}
		else if (e.equals(INTEGRITY_E_AUTH_SHA1_96)) {
			result = Encryption.getSHA1Integrity(RB, p_message, 96);
		}
		else if (e.equals(INTEGRITY_E_AUTH_HMACSHA1_64)) {
			result = Encryption.getEncryptHMACx(RB, "HmacSHA1", p_key, p_message, 64);
		}
		else {
			RB.tciTMProvided.tciError("fx_integrity: unknown algorithm " + e + "!");
			return null;
		}
//		System.out.println("fx_integrity: output: " + result.getString());
		return result;
	}
	
	/**
	 * @desc    This external function calculates the length of an IKE message
	 *			based on the length of the payload(s), length of message is 
	 *          length of payload(s) plus length of IKE header which is 28
	 * @param   p_ikepayload payload portion of IKE message
	 * @return  length of IKE message in bytes
	*/
//	TODO check if it is used	
	public synchronized IntegerValue fx_ikePayloadListLength(RecordOfValue p_ikepayload ) {
		IntegerValue ikepayloadLength = (IntegerValue)RB.tciCDRequired.getInteger().newInstance();
		//TODO
		//14 + x?
	    TriMessage message = RB.getTciEncoding().getCodec("LibIpv6_Rfc4306Ikev2_TypesAndValues").encode(p_ikepayload);
	    
	    int msgLen = message.getEncodedMessage().length/8-1;

	    ikepayloadLength.setInt(msgLen);
	    
		return ikepayloadLength;
	}
	
	/**
	 * @desc	This external function calculates the checksum for any
	 *         	UDP packet which contains an IKE message.
	 * @param  p_ikemessage IKE message 
	 * @return checksum value 
	*/
//	TODO check if it is used	
	public synchronized OctetstringValue fx_udpikeChecksum(RecordValue p_ikemessage) {
		return calcChecksum(p_ikemessage, 17);
	}

	///////////////////////////
	//inner support functions//
	///////////////////////////
	
	/**
	 * @param p_packet
	 * @param nextHdr The next header. To "skip" other headers.
	 * @return
	 */
	private OctetstringValue calcChecksum(RecordValue p_packet, int nextHdr) {
		byte [] pseudoHeader = new byte[40];
	    
	    int eh_size = 0;
	    
	    int checksum = 0;
	    
	    boolean homeAddressIncluded = false;
	    byte [] homeAddress = null;
	    
	    TabularIpv6Header ipv6Header = new TabularIpv6Header(p_packet.getField("ipv6Hdr"));
	    try {
            ipv6Header.encode(new TabularEncoder(new BitArrayOutputStream(), RB));
        } catch (TabularException tex) {
            RB.getTciTMProvided().tciError(tex.getMessage());
            tex.printStackTrace();
            return (OctetstringValue)RB.getTciCDRequired().getOctetstring().newInstance();
        }
	    
	    int nextHeader = ipv6Header.nextHeader.tValue;
        
        if (nextHeader != nextHdr) {
	        RecordOfValue ehrov = (RecordOfValue)p_packet.getField("extHdrList");
	        
	        if (!ehrov.notPresent()) {
                for (int i=0; i<ehrov.getLength(); i++) {
	                RecordValue rv = (RecordValue) ((UnionValue)ehrov.getField(i)).getVariant(((UnionValue)ehrov.getField(i)).getPresentVariantName());

	    	        switch (nextHeader) {
	    	        case HOP:
	    	            eh_size += (((IntegerValue)rv.getField("hdrExtLen")).getInt() * 8 + 8);
	    	            nextHeader = ((IntegerValue)rv.getField("nextHeader")).getInt();
	    	            break;
	    	        case TUNNELEDIPV6:
	    	            eh_size += ipv6HeaderSize;
	    	            nextHeader = ((IntegerValue)rv.getField("nextHeader")).getInt();
	    	            break;
	    	        case ROUTING:
	    	            eh_size += (((IntegerValue)rv.getField("hdrExtLen")).getInt() * 8 + 8);
	    	            nextHeader = ((IntegerValue)rv.getField("nextHeader")).getInt();
	    	            break;
	    	        case FRAGMENT:
	    	            eh_size += 8;
	    	            nextHeader = ((IntegerValue)rv.getField("nextHeader")).getInt();
	    	            break;
	    	        case DESTINATION:
	    	            eh_size += (((IntegerValue)rv.getField("hdrExtLen")).getInt() * 8 + 8);
	    	            nextHeader = ((IntegerValue)rv.getField("nextHeader")).getInt();
	    	            RecordOfValue rov = (RecordOfValue)rv.getField("destOptionList");
	    	            for (int j=0; j<rov.getLength(); j++) {
	    	            	UnionValue destOption = (UnionValue)rov.getField(j);
							Value variant = destOption.getVariant(destOption.getPresentVariantName());
							if (destOption.getPresentVariantName().equals("homeAddressOption")) {
	    	            		homeAddressIncluded = true;
	    	            		OctetstringValue ov = (OctetstringValue)((RecordValue)variant).getField("homeAddress");
	    	            		homeAddress = new OctetString(RB, ov.getString()).toByteArray();
	    	            	}
	    	            }
	    	            break;
	    	        }
	            }
	        }
	        
        }

        byte []src = null;
        if (!homeAddressIncluded) {
        	src = ipv6Header.sourceAddress.tValue;
        }
        else {
        	src = homeAddress;
        }
        System.arraycopy(src, 0, pseudoHeader, 0, 16);
        //16Byte dst
        byte []dst = ipv6Header.destinationAddress.tValue;
        System.arraycopy(dst, 0, pseudoHeader, 16, 16);

        //4Byte upper layer packet length
        byte[] upperLayerLength = new byte[4];
        int len = ipv6Header.payloadLength.tValue - eh_size;
        upperLayerLength[0] = (byte)(len>>24);
        upperLayerLength[1] = (byte)(len>>16);
        upperLayerLength[2] = (byte)(len>>8);
        upperLayerLength[3] = (byte)(len);
        System.arraycopy(upperLayerLength, 0, pseudoHeader, 32, 4);

        //3Byte zero
        byte [] zero_bytes = new byte[] {0x0, 0x0, 0x0};
        System.arraycopy(zero_bytes, 0, pseudoHeader, 36, 3);

        //1Byte next header
        byte [] next_header = new byte[] {(byte)nextHeader };
        System.arraycopy(next_header, 0, pseudoHeader, 39, 1);

        //checksum pseudo header
        byte []val = new byte[2];
        for (int i=0; i<40; i+=2) {
        	System.arraycopy(pseudoHeader, i, val, 0, 2);
        	checksum += new BigInteger(val).intValue()&0x0000FFFF;
        	checksum = (checksum>>16) + (checksum & 0x0000FFFF);
        }

        //checksum layer 4 header
//        byte [] wholeMessage = RB.getTciEncoding().getCodec("LibIpv6_Interface_TypesAndValues").encode(p_packet).getEncodedMessage();
//        for (int i=0; i<len; i+=2) {
//        	if ((i+2)<=len) {
//        		System.arraycopy(wholeMessage, i + ipv6HeaderSize + eh_size, val, 0, 2);
//        		checksum += new BigInteger(val).intValue()&0x0000FFFF;
//        		checksum = (checksum>>16) + (checksum & 0x0000FFFF);
//        	}
//        }
//
//        if ((len%2) != 0) {
//        	byte v1 = wholeMessage[len - 1 + ipv6HeaderSize + eh_size];
//        	val[0] = v1;
//        	val[1] = 0x0;
//
//        	checksum += new BigInteger(val).intValue()&0x0000FFFF;
//        	checksum = (checksum>>16) + (checksum & 0x0000FFFF);
//        }
//
//        checksum = (~checksum) & 0x0000FFFF;
//
//	    OctetstringValue ret = (OctetstringValue) RB.getTciCDRequired().getTypeForName("LibCommon_DataStrings.Oct2").newInstance();
//	    ret.setLength(2);
//	    ret.setOctet(0, (int)checksum>>8);
//	    ret.setOctet(1, (int)checksum&0xFF);
//	    
//	    return ret;
        return checksum16BitComplement(
        		(OctetstringValue) RB.getTciCDRequired().getTypeForName("LibCommon_DataStrings.Oct2").newInstance(), 
        		RB.getTciEncoding().getCodec("LibIpv6_Interface_TypesAndValues").encode(p_packet).getEncodedMessage(), 
        		checksum, 
        		ipv6HeaderSize + eh_size, len);
	}
	
	/**
	 * @param e
	 * @param p_key
	 * @param p_message
	 * @return
	 */
	//public due use in com.testingtech.ttcn.tci.codec.ike.EncryptedPayloadCodec
	public OctetstringValue integrity(String e, OctetstringValue p_key, OctetstringValue p_message) {
		OctetstringValue result = null;
		
		if (e.equals(INTEGRITY_E_AUTH_NULL)) {
			result = p_message;
		}
		else if (e.equals(INTEGRITY_E_AUTH_HMACSHA1_96)) {
			result = Encryption.getEncryptHMACx(RB, "HmacSHA1", p_key, p_message, 96);
		}
		else if (e.equals(INTEGRITY_E_AUTH_AESXCBC_MAC_96)) {
			RB.tciTMProvided.tciError("fx_integrity: algorithm " + e + " not yet implemented!");
			result = null;
		}
		else if (e.equals(INTEGRITY_E_AUTH_HMACMD5_96)) {
			result = Encryption.getEncryptHMACx(RB, "HmacMD5", p_key, p_message, 96);
		}
		else if (e.equals(INTEGRITY_E_AUTH_SHA1)) {
			result = Encryption.getSHA1Integrity(RB, p_message, 160);
		}
		else if (e.equals(INTEGRITY_E_AUTH_SHA1_96)) {
			result = Encryption.getSHA1Integrity(RB, p_message, 96);
		}
		else if (e.equals(INTEGRITY_E_AUTH_HMACSHA1_64)) {
			result = Encryption.getEncryptHMACx(RB, "HmacSHA1", p_key, p_message, 64);
		}
		else {
			RB.tciTMProvided.tciError("fx_mac: unknown algorithm " + e + "!");
			return null;
		}
//		System.out.println("fx_mac: output: " + result.getString());
		return result;
	}

	/**
	 * @param p_first
	 * @param p_second
	 * @param algo
	 * @param prv
	 * @return
	 */
	private byte[] innerPRF(String algo, byte [] p_first, byte [] p_second) {
		if (algo.equals(PRF_E_RESERVED)) {
			return p_second;
		}
		else if (algo.equals(PRF_E_HMAC_MD5)) {
			return ((OctetString) Encryption.getEncryptHMACx(RB, "HmacMD5",
					new OctetString(RB, p_first.length, p_first),
					new OctetString(RB, p_second.length, p_second), 160))
					.toByteArray();
		}
		else if (algo.equals(PRF_E_HMAC_SHA1)) {
			return ((OctetString) Encryption.getEncryptHMACx(RB, "HmacSHA1",
					new OctetString(RB, p_first.length, p_first),
					new OctetString(RB, p_second.length, p_second), 160))
					.toByteArray();
		}
		else if (algo.equals(PRF_E_HMAC_TIGER)) {
			RB.tciTMProvided.tciError("fx_pseudoRandom: " + algo + " not yet implemented!");
			return null;
		}
		else if (algo.equals(PRF_E_AES_128_XCBC)) {
			RB.tciTMProvided.tciError("fx_pseudoRandom: " + algo + " not yet implemented!");
			return null;
		}
		else {
			RB.tciTMProvided.tciError("fx_pseudoRandom: unknown algorithm " + algo + "!");
			return null;
		}
	}

	private static byte [] block = null;
	private static int blockPointer = 0;
	
	/**
	 * @param cnt
	 * @param algo
	 * @param first
	 * @param second
	 * @param key
	 * @param keySize
	 * @return
	 * @throws IOException
	 */
	private int getKey(int cnt, String algo, byte [] first, byte [] second, ByteArrayOutputStream key, int keySize) throws IOException {
		//first fill the rest of the last block, if there is one
		if ((blockPointer != 0) && (block != null)) {
			int len = block.length-blockPointer;
			//if the key size is less than the available block, write only the amount of key size
			if (len>keySize) {
				key.write(block, blockPointer, keySize);
				blockPointer += keySize;
				return cnt;
			}
			//else write the full block and go on
			key.write(block, blockPointer, len);
			keySize -= len;
		}
		//get a new block
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		if (block != null) {
			baos.write(block);
		}
		baos.write(second);
		baos.write(++cnt);
		//generate a new block
		block = innerPRF(algo, first, baos.toByteArray());
		//reset the blockPointer
		blockPointer = 0;
		if (block.length < keySize) {
			key.write(block, 0, block.length);
			blockPointer = 0;
			cnt = getKey(cnt, algo, first, second, key, keySize-block.length);
		}
		else {
			key.write(block, blockPointer, keySize);
			blockPointer = keySize%block.length;
		}
		
		return cnt;
	}
	
//	private int getKeySize(String algo) {
//		if (algo.equals(INTEGRITY_E_AUTH_NULL) || algo.equals(ENCRYPTION_E_NULL) || algo.equals(INTEGRITY_E_NONE) || algo.equals(ENCRYPTION_E_RESERVED)) {
//			return 0;
//		}
//		else if (algo.equals(INTEGRITY_E_AUTH_AESXCBC_MAC_96) || algo.equals(INTEGRITY_E_AUTH_AESXCBC_96)) {
//			return 16;
//		}
//		else if (algo.equals(PRF_E_HMAC_MD5) || algo.equals(INTEGRITY_E_AUTH_HMACMD5_96)) {
//			return 16;
//		}
//		else if (algo.equals(INTEGRITY_E_AUTH_HMACSHA1_64)) {
//			return 20;
//		}
//		else if (algo.equals(PRF_E_HMAC_SHA1) || algo.equals(INTEGRITY_E_AUTH_HMACSHA1_96)) {
//			return 20;
//		}
//		else if (algo.equals(INTEGRITY_E_AUTH_SHA1)) {
//			return 0;
//		}
//		else if (algo.equals(INTEGRITY_E_AUTH_SHA1_96)) {
//			return 0;
//		}
//		else if (algo.equals(ENCRYPTION_E_AES_CBC)) {
//			return ((IPv6TestAdapter)RB.TestAdapter).getIkeEncryptionKeyLength();
//		}
//		else if (algo.equals(ENCRYPTION_E_AES_CTR)) {
//			return ((IPv6TestAdapter)RB.TestAdapter).getIkeEncryptionKeyLength();
//		}
//		else if (algo.equals(ENCRYPTION_E_DES_CBC)) {
//			return 8;
//		}
//		else if (algo.equals(ENCRYPTION_E_TRIPLEDES_CBC)) {
//			return 24;
//		}
//		else if (algo.equals(INTEGRITY_E_AUTH_DESMAC)) {
//			//TODO no info so far
//			return 0;
//		}
//		else if (algo.equals(INTEGRITY_E_AUTH_KPDKMD5)) {
//			//TODO no info so far
//			return 0;
//		}
//		else if (algo.equals(ENCRYPTION_E_DES_IV_64)) {
//			//TODO no info so far
//			return 0;
//		}
//		else if (algo.equals(ENCRYPTION_E_RC5)) {
//			return 16;
//		}
//		else if (algo.equals(ENCRYPTION_E_IDEA)) {
//			return 16;
//		}
//		else if (algo.equals(ENCRYPTION_E_CAST)) {
//			return ((IPv6TestAdapter)RB.TestAdapter).getIkeEncryptionKeyLength();
//		}
//		else if (algo.equals(ENCRYPTION_E_BLOWFISH)) {
//			return ((IPv6TestAdapter)RB.TestAdapter).getIkeEncryptionKeyLength();
//		}
//		else if (algo.equals(ENCRYPTION_E_3IDEA)) {
//			//TODO to be confirmed
//			return 48;
//		}
//		else if (algo.equals(ENCRYPTION_E_DES_IV_32)) {
//			//TODO no info so far
//			return 0;
//		}
//		else {
//			RB.tciTMProvided.tciError("Unknown algorithm size of " + algo);
//			return 0;
//		}
//	}
		
	/**
	 * Calculates the Diffie-Hellman keys.
	 * @param p_diffieHellmanGroup
	 * @param p_diffieHellmanPrivKey
	 * @return
	 */
	private OctetstringValue calcDH(EnumeratedValue p_diffieHellmanGroup, OctetstringValue p_diffieHellmanPrivKey, BigInteger mat) {
		BigInteger dh = null;
		if (p_diffieHellmanGroup.getEnum().equals(DH_E_NONE)) {
			return p_diffieHellmanPrivKey;
		}
		else if (p_diffieHellmanGroup.getEnum().equals(DH_E_GROUP1_MODP768BIT)) {
			dh = PRIME_DH_GROUP1_MODP768BIT;
		}
		else if (p_diffieHellmanGroup.getEnum().equals(DH_E_GROUP2_MODP1024BIT)) {
			dh = PRIME_DH_GROUP1_MODP1024BIT;
		}
		else if (p_diffieHellmanGroup.getEnum().equals(DH_E_GROUP5_MODP1536BIT)) {
			dh = PRIME_DH_GROUP1_MODP1536BIT;
		}
		else if (p_diffieHellmanGroup.getEnum().equals(DH_E_GROUP14_MODP2048BIT)) {
			dh = PRIME_DH_GROUP1_MODP2048BIT;
		}
		else if (p_diffieHellmanGroup.getEnum().equals(DH_E_GROUP15_MODP3072BIT)) {
			dh = PRIME_DH_GROUP1_MODP3072BIT;
		}
		else if (p_diffieHellmanGroup.getEnum().equals(DH_E_GROUP16_MODP4096BIT)) {
			dh = PRIME_DH_GROUP1_MODP4096BIT;
		}
		else if (p_diffieHellmanGroup.getEnum().equals(DH_E_GROUP17_MODP6144BIT)) {
			dh = PRIME_DH_GROUP1_MODP6144BIT;
		}
		else if (p_diffieHellmanGroup.getEnum().equals(DH_E_GROUP18_MODP8192BIT)) {
			dh = PRIME_DH_GROUP1_MODP8192BIT;
		}
		else {
			RB.tciTMProvided.tciError(p_diffieHellmanGroup.getEnum() + " is not implemented in the external function.");
			return null;
		}

		return new BigInt(RB, mat.modPow(new BigInteger(1,
				((OctetString) p_diffieHellmanPrivKey).toByteArray()), dh))
		.int2oct();
	}

	private static final String[] hex = { "0123456789abcdef", "0123456789ABCDEF" };

	/**
	 * This external function calculates the total length
	 * of a IPv4 packet.
	 * 
	 * @param   p_msg Ipv4 packet
	 * @return  payload length in bytes
	 */
	public synchronized IntegerValue fx_ipv4PacketTotalLength(RecordValue p_msg ){
	    
	    TriMessage message = RB.getTciEncoding().getCodec("LibIpv6_Interface_TypesAndValues").encode(p_msg);
	    
	    int msgLen = message.getEncodedMessage().length;
	    
	    IntegerValue ret = (IntegerValue) RB.getTciCDRequired().getTypeForName("LibCommon_BasicTypesAndValues.UInt16").newInstance();

	    ret.setInt(msgLen);
	    
	    return ret;
	}

	/**
	 * Calculates the checksum over the IPv4 payload
	 * 
	 * @param  p_payload Checksum is calculated over this
	 * @return checksum value
	 */
	public synchronized OctetstringValue fx_calcIpv4PayloadChecksum(UnionValue p_payload) {
	    OctetstringValue ret = (OctetstringValue) RB.getTciCDRequired().getTypeForName("LibCommon_DataStrings.Oct2").newInstance();
	    
        if (p_payload.getPresentVariantName().equals("ipv6Packet")) {
        	RB.getTciTLProvided()
					.tliRT("", System.nanoTime(), "", -1, null,
							RTLoggingConstants.RT_LOG_ERROR,
							"No Ipv4 payload checksum calculation needed for ipv6Packet!");
        	return ret;
        }
        
        RecordValue rv = (RecordValue) p_payload.getVariant(p_payload.getPresentVariantName());

		TabularEncoder enc = new TabularEncoder(new BitArrayOutputStream());
		
		try {
			//icmpType
			enc.encodeInteger(((IntegerValue)rv.getField("icmpType")).getInt(), 8);
			//icmpCode
			enc.encodeInteger(((IntegerValue)rv.getField("icmpCode")).getInt(), 8);
			//checksum: zero
			enc.encodeOctetstring(new byte []{(byte) 0x00, (byte) 0x00});
			//identifier
			enc.encodeInteger(((IntegerValue)rv.getField("identifier")).getInt(), 16);
			//sequenceNumber
			enc.encodeInteger(((IntegerValue)rv.getField("sequenceNumber")).getInt(), 16);
			//data
			if (!rv.getField("data").notPresent()) {
				enc.encodeOctetstring(((OctetString)rv.getField("data")).toByteArray());
			}	
		} catch (TabularException e) {
        	RB.getTciTLProvided().tliRT(
					"",
					System.nanoTime(),
					"",
					-1,
					null,
					RTLoggingConstants.RT_LOG_ERROR,
					"fx_calcIpv4PayloadChecksum produces following error: "
							+ e.getMessage());
        	return ret;
		}		
		
        byte[] wholeMessage = enc.toByteArray();
		//checksum layer 4 header
        return checksum16BitComplement(ret, wholeMessage, 0, 0, wholeMessage.length);
	}

	/**
	 * Calculates the header checksum (HCS) over the IPv4 header
	 * 
	 * @param  p_ipv4Header Checksum is calculated over this
	 * @return checksum value
	 */
	public synchronized OctetstringValue fx_calcIpv4Hcs(RecordValue p_ipv4Header) {
		OctetstringValue ret = (OctetstringValue) RB.getTciCDRequired().getTypeForName("LibCommon_DataStrings.Oct2").newInstance();
	    
		TabularEncoder enc = new TabularEncoder(new BitArrayOutputStream());
		
		try {
			//version
			enc.encodeInteger(((IntegerValue)p_ipv4Header.getField("version")).getInt(), 4);
			//headerLength
			enc.encodeInteger(((IntegerValue)p_ipv4Header.getField("headerLength")).getInt(), 4);
			//typeOfService
			enc.encodeBitstring(((BitString)p_ipv4Header.getField("typeOfService")).toByteArray(), 8);
			//totalLength
			enc.encodeInteger(((IntegerValue)p_ipv4Header.getField("totalLength")).getInt(), 16);
			//identification
			enc.encodeOctetstring(((OctetString)p_ipv4Header.getField("identification")).toByteArray());
			//reserved
			enc.encodeInteger(((IntegerValue)p_ipv4Header.getField("reserved")).getInt(), 1);
			//doNotFragFlag
			enc.encodeInteger(((IntegerValue)p_ipv4Header.getField("doNotFragFlag")).getInt(), 1);
			//moreFragsFlag
			enc.encodeInteger(((IntegerValue)p_ipv4Header.getField("moreFragsFlag")).getInt(), 1);
			//fragmentOffset
			enc.encodeInteger(((IntegerValue)p_ipv4Header.getField("fragmentOffset")).getInt(), 13);
			//timeToLive
			enc.encodeInteger(((IntegerValue)p_ipv4Header.getField("timeToLive")).getInt(), 8);
			//protocol
			enc.encodeInteger(((IntegerValue)p_ipv4Header.getField("protocol")).getInt(), 8);
			//hcs: zero
			enc.encodeOctetstring(new byte []{(byte) 0x00, (byte) 0x00});
			//sourceAddress
			enc.encodeOctetstring(((OctetString)p_ipv4Header.getField("sourceAddress")).toByteArray());
			//destinationAddress
			enc.encodeOctetstring(((OctetString)p_ipv4Header.getField("destinationAddress")).toByteArray());
			//ipv4HdrOptions
			if (!p_ipv4Header.getField("ipv4HdrOptions").notPresent()) {
				enc.encodeOctetstring(((OctetString)p_ipv4Header.getField("ipv4HdrOptions")).toByteArray());
			}	
		} catch (TabularException e) {
        	RB.getTciTLProvided().tliRT(
					"",
					System.nanoTime(),
					"",
					-1,
					null,
					RTLoggingConstants.RT_LOG_ERROR,
					"fx_calcIpv4PayloadChecksum produces following error: "
							+ e.getMessage());
        	return ret;
		}		
		
        byte[] wholeMessage = enc.toByteArray();
		//checksum layer 4 header
        return checksum16BitComplement(ret, wholeMessage, 0, 0, wholeMessage.length);
	}

	/**
	 * @param ret
	 * @param wholeMessage
	 * @param checksum TODO
	 * @param offset TODO
	 * @param len TODO
	 * @return
	 */
	private OctetstringValue checksum16BitComplement(OctetstringValue ret, byte[] wholeMessage, int checksum, int offset, int len) {        
        byte []val = new byte[2];

        for (int i=0; i<len; i+=2) {
        	if ((i+2)<=len) {
        		System.arraycopy(wholeMessage, i + offset, val, 0, 2);
        		checksum += new BigInteger(val).intValue()&0x0000FFFF;
        		checksum = (checksum>>16) + (checksum & 0x0000FFFF);
        	}
        }

        if ((len%2) != 0) {
        	byte v1 = wholeMessage[len - 1 + offset];
        	val[0] = v1;
        	val[1] = 0x0;

        	checksum += new BigInteger(val).intValue()&0x0000FFFF;
        	checksum = (checksum>>16) + (checksum & 0x0000FFFF);
        }

        checksum = (~checksum) & 0x0000FFFF;

	    ret.setLength(2);
	    ret.setOctet(0, (int)checksum>>8);
	    ret.setOctet(1, (int)checksum&0xFF);
	    
	    return ret;
	}

}
